LocalConfigData = {
    "env_select":"simple_river",
    "adapter_select": ["Default", "Language"],
    "training_action_cap": 100,
    "testing_action_cap":100,
    "reward_signal": [0.5,0,-0.1],
    "sub_goal": "None",
    "supervised_rewards":"False",
    "y_limit":25,
    "obs_precision":2
}


from datetime import datetime
import os
import urllib.request
import json 
import httpimport
# Local imports
from elsciRL.application_suite.import_data import Applications
from elsciRL.application_suite.search_agent import DefaultAgentConfig

class PullApplications:
    """Simple applications class to run a setup tests of experiments.
        - Problem selection: problems to run in format ['problem1', 'problem2',...]

    Applications:
        - Sailing: {'easy'},
        - Classroom: {'classroom_A'}
    """
    # TODO: Make it so it pulls all possible configs and adapters from the repo
    def __init__(self) -> None:
        imports = Applications()
        self.imports = imports.data
        self.current_test = {}
        
        
    def pull(self, problem_selection:list=[]):
        # Pull all problems if none are selected
        if len(problem_selection)>0:
            self.problem_selection = problem_selection
        else:
            self.problem_selection = list(self.imports.keys())
        
        # Extract data from imports
        for problem in list(self.problem_selection):
            engine = self.imports[problem]['engine_filename']
            if problem not in self.imports:
                raise ValueError(f"Problem {problem} not found in the setup tests.")
            else:
                self.current_test[problem] = {}
                # current_test = {'problem1': {'engine':engine.py, 'local_configs': {'config1':config.json, 'config2':config.json}, 'adapters': {'adapter1':adapter.py, 'adapter2':adapter.py}}}
                root = 'https://raw.githubusercontent.com/'+ self.imports[problem]['github_user'] + "/" + self.imports[problem]['repository'] + "/" + self.imports[problem]['commit_id']
                # NOTE - This requires repo to match structure with engine inside environment folder
                engine_module = httpimport.load(engine, root+'/'+self.imports[problem]['engine_folder']) 
                # TODO: Pull class name directly from engine file to be called
                self.current_test[problem]['engine'] = engine_module.Engine
            # ------------------------------------------------
            # - Pull Adapters, Configs and Analysis
            self.current_test[problem]['adapters'] = {}
            for adapter_name, adapter in self.imports[problem]['adapter_filenames'].items():
                adapter_module = httpimport.load(adapter, root+'/'+self.imports[problem]['local_adapter_folder'])   
                # TODO: Pull class name directly from adapter file to be called
                try:
                    self.current_test[problem]['adapters'][adapter_name] = adapter_module.DefaultAdapter
                except:
                    self.current_test[problem]['adapters'][adapter_name] = adapter_module.LanguageAdapter
            # ---
            self.current_test[problem]['experiment_configs'] = {}
            for config_name,config in self.imports[problem]['experiment_config_filenames'].items():
                experiment_config = json.loads(urllib.request.urlopen(root+'/'+self.imports[problem]['config_folder']+'/'+config).read())
                self.current_test[problem]['experiment_configs'][config_name] = experiment_config
            # ---
            self.current_test[problem]['local_configs'] = {}
            for config_name,config in self.imports[problem]['local_config_filenames'].items():
                local_config = json.loads(urllib.request.urlopen(root+'/'+self.imports[problem]['config_folder']+'/'+config).read())
                self.current_test[problem]['local_configs'][config_name] = local_config
            # ---
            self.current_test[problem]['local_analysis'] = {}
            for analysis_name,analysis in self.imports[problem]['local_analysis_filenames'].items():
                try:
                    local_analysis = httpimport.load(analysis, root+'/'+self.imports[problem]['local_analysis_folder'])  
                    # TODO: Pull class name directly from analysis file to be called 
                    self.current_test[problem]['local_analysis'][analysis_name] = local_analysis.Analysis
                except:
                    print("No analysis file found.")
                    self.current_test[problem]['local_analysis'][analysis_name] = {}
            
            # ------------------------------------------------
            # Pull prerender data
            print("-----------------------------------------------")
            print(problem)
            print("Source: ", root)
            self.current_test[problem]['prerender_data'] = {}
            if self.imports[problem]['prerender_data_folder'] != '':
                try:
                    for prerender_name, prerender in self.imports[problem]['prerender_data_filenames'].items():
                        if prerender.endswith(('.txt', '.json', '.xml')):
                            data = json.loads(urllib.request.urlopen(root+'/'+self.imports[problem]['prerender_data_folder']+'/'+prerender).read().decode('utf-8'))
                            self.current_test[problem]['prerender_data'][prerender_name] = data
                    print("Pulling prerender data...")
                except:
                    print(root+'/'+self.imports[problem]['prerender_data_folder']+'/'+prerender)
                    print("No prerender data found.")
                    self.current_test[problem]['prerender_data'] = {}
            else:
                print("No prerender data found.")
                self.current_test[problem]['prerender_data'] = {}

            # Pull prerender images
            self.current_test[problem]['prerender_images'] = {}
            if self.imports[problem]['prerender_data_folder'] != '':
                try:
                    for image_name, image in self.imports[problem]['prerender_image_filenames'].items():
                        if image.endswith(('.png', '.jpg', '.svg', '.gif')):
                            image_url = root + '/' + self.imports[problem]['prerender_data_folder'] + '/' + image
                            image_data = urllib.request.urlopen(image_url).read()
                            self.current_test[problem]['prerender_images'][image_name] = image_data
                    print("Pulling prerender images...")
                except:
                    print("No prerender images found.")
                    self.current_test[problem]['prerender_images'] = {}
            else:
                print("No prerender images found.")
                self.current_test[problem]['prerender_images'] = {}
            # -----------------------------------------------
        print("-----------------------------------------------")

        return self.current_test

    def setup(self, agent_config:dict={}) -> None:
        if agent_config == {}:
            agent_config = DefaultAgentConfig()
            self.ExperimentConfig = agent_config.data  
        else:
            self.ExperimentConfig = agent_config 

        return self.ExperimentConfig
    
    def add_applicaiton(self, problem:str, application_data:dict) -> None:
        """ Add a new application to the list of applications. 
        Reqired form:
            - engine: <engine.py>
            - experiment_configs: {experiment_config:experiment_config_path|<experiment_config.json>}
            - local_configs: {local_config:experiment_config_path|<local_config.json>}
            - adapters: {adapter:<adapter.py>}
            - local_analysis: {analysis:<analysis.py>}
            - prerender_data: {data:data_path|<data.txt>}
            - prerender_images: {image:<image.png>}
        """
        # ---
        # Get configs and data from path directory to imported json/txt files
        if type(list(application_data['experiment_configs'].values())[0])==str:
            experiment_config = {}
            for name,experiment_config_dir in application_data['experiment_configs'].items():
                with open (experiment_config_dir, 'r') as f:
                    # Load the JSON data from the file
                    agent_config = json.loads(f.read())
                    experiment_config[name] = agent_config
            application_data['experiment_configs'] = experiment_config

        if type(list(application_data['local_configs'].values())[0])==str:
            local_config = {}
            for name,local_config_dir in application_data['local_configs'].items():
                with open (local_config_dir, 'r') as f:
                    # Load the JSON data from the file
                    agent_config = json.loads(f.read())
                    local_config[name] = agent_config
            application_data['local_configs'] = local_config

        if len(application_data['prerender_data'])>0:
            if type(list(application_data['prerender_data'].values())[0])==str:
                data = {}
                for name,data_dir in application_data['prerender_data'].items():
                    with open (data_dir, 'r') as f:
                        # Load the JSON data from the file
                        agent_config = json.loads(f.read().decode('utf-8'))
                        data[name] = agent_config
                application_data['prerender_data'] = data
        # ---
        self.imports[problem] = application_data
        self.current_test[problem] = application_data
        print(f"Added {problem} to the list of applications.")
        print(f"Current applications: {self.imports.keys()}")

        return self.current_test
    
    def remove_application(self, problem:str) -> None:
        # Remove an application from the list of applications
        if problem in self.imports:
            del self.imports[problem]
            del self.current_test[problem]
            print(f"Removed {problem} from the list of applications.")
        else:
            print(f"{problem} not found in the list of applications.")

        return self.imports

            


import numpy as np
import pandas as pd
from typing import List, Dict
from abc import ABC, abstractmethod
from torch import Tensor

class Encoder(ABC):
    @abstractmethod
    def encode(self, *args, **kwargs) -> Tensor:
        pass

class StateEncoder(Encoder):
    tensor_cache: Dict[int, Tensor] = dict()
    tensor_cache_index: int = 0

    @staticmethod
    def cache_insert(t: Tensor):
        StateEncoder.tensor_cache[StateEncoder.tensor_cache_index] = t
        StateEncoder.tensor_cache_index += 1
    
    @staticmethod
    def cache_retrieve(offset: int, index: int):
        return StateEncoder.tensor_cache[offset + index]

        
    def encode(self, state:list = None, legal_actions:list = None, episode_action_history:str = None) -> Tensor:
        pass
    


import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os

import warnings
warnings.filterwarnings("ignore")

from elsciRL.analysis.visual_output import VisualOutput
from elsciRL.analysis.tabular_output import TabularOutput
class Evaluation:
    """Calls the Local Environment function and produces analysis reports & figures."""
    def __init__(self, window_size) -> None:
        self.window_size = window_size
        self.timeout = 10
    
    # Training Plots
    def train_report(self, training_results:pd.DataFrame, save_dir:str, show_figures:str):
        """Produces visual and tabular analysis of individual training results."""
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)
        
        VisualAnalysis = VisualOutput(training_results, save_dir, show_figures, self.window_size)
        VisualAnalysis.episode_reward_graph()
        VisualAnalysis.cumulative_reward_graph()
        VisualAnalysis.episode_reward_dist_graph()
        VisualAnalysis.average_reward_graph()
        VisualAnalysis.rolling_average_reward_graph()
        VisualAnalysis.avg_return_dist_graph()
        VisualAnalysis.number_actions_graph()
        VisualAnalysis.number_actions_dist_graph()
        VisualAnalysis.runtime_per_episode_graph()
        # Catch runtime issues for dist plot
        #VisualAnalysis.runtime_dist_graph()
        
        TableOutput = TabularOutput(training_results, save_dir)
        TableOutput.save_results()
        return training_results[training_results['episode']==np.max(training_results['episode'])]['cumulative_reward'].iloc[0]
        
    def training_variance_report(self, save_dir:str, show_figures:str):
        """Extracts individual training reports from folders in save_dir and produces summary report if agent types match."""
        # Get first file dir name after 'output' and last
        save_dir_lst = save_dir.split("/")
        env_name = ''
        flag = False
        for st in save_dir_lst:
            if flag == True:
                env_name = env_name+st+'\n'
            if st=='output':
                flag = True

        sub_folders = [name for name in os.listdir(save_dir) if os.path.isdir(os.path.join(save_dir, name))]
        sub_folders = [name for name in sub_folders if ('train' in name)]
        sub_folders.sort()
        results = {}
        current_list = []
        prior_agent_type = '_'.join(sub_folders[0].split('_')[:-1])
        for name in sub_folders:
            if ('train' in name) | ('test' in name):
                agent_type = '_'.join(name.split('_')[:-1])
            
            if agent_type not in results:
                results[agent_type] = {}

            #Generate list until type no longer matches
            if agent_type == prior_agent_type:
                current_list.append(name)
            else:
                results[prior_agent_type] = current_list
                current_list = [name]
            prior_agent_type = agent_type
        results[prior_agent_type] = current_list # needed to log final agent_type data
        
        # If path name has start position->goal in it then uses seeds, otherwise single start position repeats
        if agent_type.split("_")[-2]!='training':
            training_type = 'seeds'
        else:
            training_type = 'repeats'

        # Speed this up with dict search instead of df iteration
        #cycol = cycle('brcmk')
        cycol = 'brgcmyk'
        line_styles = ['solid','dotted','dashed','dashdot', 'solid','dotted','dashed','dashdot']
        col = 0
        fig, axs = plt.subplots(2,2)
        num_episodes_max = 0
        for agent_type in list(results.keys()): 
            combine_results = pd.DataFrame()
            final_results = pd.DataFrame()
            num_repeats=0
            for dir_path in results[agent_type]:    
                path = save_dir+'/'+dir_path+"/results.csv"
                df = pd.read_csv(path)
                # -----------------
                # Calculate delta as difference between reward obtained from prior episode
                # Segment delta into batches and average to 'smooth' results calculation        
                if (int(np.max(df['episode']*self.window_size))+1) <100:
                    w = int(np.max(df['episode']*self.window_size))+1
                else:
                    w = 100
                reward_delta = []
                for n,r in enumerate(df['episode_reward']): # End when window size no longer fits
                    if n < w:
                        window_reward = np.mean(df['episode_reward'][:n])
                    else:
                        window_reward = np.mean(df['episode_reward'][n-w:n])
                    reward_delta.append(window_reward)
                avg_r_df = pd.DataFrame({'episode':df['episode'], 'avg_R':reward_delta, 'cum_R':df['cumulative_reward'], 'time':df['time_per_episode']})
                combine_results = pd.concat([combine_results, avg_r_df], ignore_index=True)
                num_repeats+=1

            combine_results = combine_results.reset_index()
            num_episode = np.max(combine_results['episode'])
            if num_episode>num_episodes_max:
                num_episodes_max = num_episode
            for epi in range(0,(num_episode+1)):
                final_results = pd.concat([final_results, 
                                    pd.DataFrame.from_records([{ 
                                        'agent': str(agent_type),
                                        #'opponent': combine_results['opponent'].iloc[epi],
                                        'num_repeats': len(combine_results[combine_results['episode']==epi]),
                                        'episode': epi, 
                                        "avg_R_mean": np.mean(combine_results[combine_results['episode']==epi]['avg_R']),
                                        "avg_R_se": np.std(combine_results[combine_results['episode']==epi]['avg_R'], ddof=1) / np.sqrt(np.size(combine_results[combine_results['episode']==epi]['avg_R'])),
                                        "cum_R_mean": np.mean(combine_results[combine_results['episode']==epi]['cum_R']),
                                        "cum_R_se": np.std(combine_results[combine_results['episode']==epi]['cum_R'], ddof=1) / np.sqrt(np.size(combine_results[combine_results['episode']==epi]['cum_R'])),
                                        "time_mean": np.median(combine_results[combine_results['episode']==epi]['time'])
                                        }])],
                                    ignore_index=True)

            avg_r_mean_sorted = np.sort(final_results['avg_R_mean'])
            cdf_mean = 1. * np.arange(len(avg_r_mean_sorted)) / (len(avg_r_mean_sorted) - 1)

            # Plot RL Reward results for each approach
            # 1.1 Summary of total REWARD
            if col >= len(cycol):
                col = 0
            c = cycol[col]
            l = line_styles[col]
            if col <= len(cycol):
                col+=1
            else:
                c = np.random.rand(len(x),3)
                l = 'solid'                
            x =  final_results['episode']
            avg_R = np.array(final_results['avg_R_mean'])
            avg_R_SE = np.array(final_results['avg_R_se'])
            cum_R = np.array(final_results['cum_R_mean'])
            cum_R_SE = np.array(final_results['cum_R_se'])
            time_mean = np.array(final_results['time_mean'])
            
            axs[0,0].plot(x,avg_R, color=c, linestyle=l, label=agent_type)
            axs[0,0].fill_between(x,avg_R-avg_R_SE, avg_R+avg_R_SE, color=c, alpha = 0.2)
            axs[0,1].plot(avg_r_mean_sorted,cdf_mean, color=c, linestyle=l)
            axs[1,0].plot(x,cum_R, color=c, linestyle=l)
            axs[1,0].fill_between(x,cum_R-cum_R_SE, cum_R+cum_R_SE, color=c, alpha = 0.2)
            axs[1,1].hist(time_mean, color=c, alpha=0.25)
            # Save results to csv per agent type
            final_results.to_csv(save_dir+'/training_variance_results_'+str(agent_type.split('__')[0])+'.csv')

        axs[0,0].set_xlabel("Episode")
        axs[0,0].set_ylabel('Reward')
        axs[0,0].axes.get_xaxis().set_ticks([0, num_episodes_max])
        axs[0,0].set_title("Mean and Std. Err. of Rolling Avg. R (window size="+str(w)+ " epi)")
        
        axs[0,1].set_ylabel("Cumulative Probability")
        axs[0,1].set_xlabel("Mean Reward per Episode Window")
        axs[0,1].set_title("CDF of Rolling Average R")
        
        axs[1,0].set_xlabel("Episode")
        axs[1,0].set_ylabel('Cumulative Reward')
        axs[1,0].axes.get_xaxis().set_ticks([0, num_episodes_max])
        axs[1,0].set_title("Cumulative R with Std. Err.")
        
        axs[1,1].set_ylabel("Occurence")
        axs[1,1].set_xlabel("Time")
        axs[1,1].set_title("Dist of Time per Episode")

        #ax1.legend(loc=2, bbox_to_anchor=(-0.05, 0), fancybox=True, shadow=True, framealpha=1)
        #ax2.legend(loc=2, bbox_to_anchor=(0, 1), fancybox=True, shadow=True, framealpha=1)
        if training_type=='repeats':
            fig.suptitle("Training Results for: "+str(env_name)+'Variance over '+str(num_repeats)+" runs with random & independent repeats")
        else:
            fig.suptitle("Training Results for: "+str(env_name)+"with random environment seeds")
        fig.legend(loc='upper right', fancybox=True, shadow=True, framealpha=1)
        fig.set_size_inches(12, 8)
        fig.tight_layout()

        if show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            fig.savefig(save_dir+'/training_variance_comparison.png', dpi=100)
            plt.close()
        else:
            fig.savefig(save_dir+'/training_variance_comparison.png', dpi=100)
            plt.close()

 
    # Compare the performance of trained agent vs trained agent
    def test_report(self, test_results:pd.DataFrame, save_dir:str, show_figures:str):
        """Produces visual and tabular analysis of testing results."""
        VisualAnalysis = VisualOutput(test_results, save_dir, show_figures, self.window_size)
        VisualAnalysis.episode_reward_graph()
        VisualAnalysis.cumulative_reward_graph()
        VisualAnalysis.episode_reward_dist_graph()
        VisualAnalysis.average_reward_graph()
        VisualAnalysis.rolling_average_reward_graph()
        VisualAnalysis.avg_return_dist_graph()
        VisualAnalysis.number_actions_graph()
        VisualAnalysis.number_actions_dist_graph()
        VisualAnalysis.runtime_per_episode_graph()
        # Catch runtime issues for dist plot
        #VisualAnalysis.runtime_dist_graph()

        TableOutput = TabularOutput(test_results, save_dir)
        TableOutput.save_results()
        return test_results[test_results['episode']==np.max(test_results['episode'])]['cumulative_reward'].iloc[0]

    def testing_variance_report(self, save_dir:str, show_figures:str):
        """Extracts individual training reports from folders in save_dir and produces summary report if agent types match."""
        # Get first file dir name after 'output' and last
        save_dir_lst = save_dir.split("/")
        env_name = ''
        flag = False
        for st in save_dir_lst:
            if flag == True:
                env_name = env_name+st+'\n'
            if st=='output':
                flag = True

        sub_folders = [name for name in os.listdir(save_dir) if os.path.isdir(os.path.join(save_dir, name))]
        sub_folders = [name for name in sub_folders if ('train' in name)|('test' in name)]
        sub_folders = [name for name in sub_folders if name.split("__")[1].split("_")[0]=='testing']
        print(sub_folders)
        sub_folders.sort()
        results = {}
        current_list = []
        prior_agent_type = ''.join(sub_folders[0].split('__')[0])
        
        for name in sub_folders:
            agent_type = ''.join(name.split('__')[0])
            if agent_type not in results:
                results[agent_type] = {}

             #Generate list until type no longer matches
            if agent_type == prior_agent_type:
                current_list.append(name)
            else:
                results[prior_agent_type] = current_list
                current_list = [name]
            prior_agent_type = agent_type
        results[prior_agent_type] = current_list  # needed to log final agent_type data
        print(results)
        # Speed this up with dict search instead of df iteration
        cycol = 'brgcmyk'
        line_styles = ['solid','dotted','dashed','dashdot', 'solid','dotted','dashed','dashdot']
        col = 0
        fig, axs = plt.subplots(2,2)
        for agent_type in list(results.keys()): 
            combine_results = pd.DataFrame()
            final_results = pd.DataFrame()
            num_repeats=0
            test_folders = results[agent_type]
            #for dir_path in results[agent_type]:  
                #test_folders = [name for name in os.listdir(save_dir+'/'+dir_path) if os.path.isdir(os.path.join(save_dir+'/'+dir_path, name))]
            print(test_folders)
            for testing_path in test_folders:  
                path = save_dir+"/"+testing_path+"/results.csv"
                print(path)
                df = pd.read_csv(path)
                # -----------------
                # Calculate delta as difference between reward obtained from prior episode
                # Segment delta into batches and average to 'smooth' results calculation        
                if (int(np.max(df['episode']*self.window_size))+1) <100:
                    w = int(np.max(df['episode']*self.window_size))+1
                else:
                    w = 100
                reward_delta = []
                for n,r in enumerate(df['episode_reward']): # End when window size no longer fits
                    # At start number of prior episodes is less than window size
                    if n < w:
                        window_reward = np.mean(df['episode_reward'][:n])
                    else:
                        window_reward = np.mean(df['episode_reward'][n-w:n])
                    reward_delta.append(window_reward)
                avg_r_df = pd.DataFrame({'episode':df['episode'], 'avg_R':reward_delta, 'cum_R':df['cumulative_reward'], 'time':df['time_per_episode']})
                combine_results = pd.concat([combine_results, avg_r_df], ignore_index=True)
                num_repeats+=1

            combine_results = combine_results.reset_index()
            num_episode = np.max(combine_results['episode'])
            for epi in range(0,(num_episode+1)):
                final_results = pd.concat([final_results, 
                                    pd.DataFrame.from_records([{ 
                                        'agent': str(agent_type),
                                        #'opponent': combine_results['opponent'].iloc[epi],
                                        'num_repeats': len(combine_results[combine_results['episode']==epi]),
                                        'episode': epi, 
                                        "avg_R_mean": np.mean(combine_results[combine_results['episode']==epi]['avg_R']),
                                        "avg_R_se":  np.std(combine_results[combine_results['episode']==epi]['avg_R'], ddof=1) / np.sqrt(np.size(combine_results[combine_results['episode']==epi]['avg_R'])),
                                        "cum_R_mean": np.mean(combine_results[combine_results['episode']==epi]['cum_R']),
                                        "cum_R_se": np.std(combine_results[combine_results['episode']==epi]['cum_R'], ddof=1) / np.sqrt(np.size(combine_results[combine_results['episode']==epi]['cum_R'])),
                                        "time_mean": np.median(combine_results[combine_results['episode']==epi]['time'])
                                     }])],
                                    ignore_index=True)
            
            avg_r_mean_sorted = np.sort(final_results['avg_R_mean'])
            cdf_mean = 1. * np.arange(len(avg_r_mean_sorted)) / (len(avg_r_mean_sorted) - 1)

            # Plot RL Reward results for each approach
            # 1.1 Summary of total REWARD
            if col >= len(cycol):
                col = 0
            c = cycol[col]
            l = line_styles[col]
            if col <= len(cycol):
                col+=1
            else:
                c = np.random.rand(len(x),3)
                l = 'solid'

            x =  final_results['episode']
            avg_R = np.array(final_results['avg_R_mean'])
            avg_R_SE = np.array(final_results['avg_R_se'])
            cum_R = np.array(final_results['cum_R_mean'])
            cum_R_SE = np.array(final_results['cum_R_se'])
            time_mean = np.array(final_results['time_mean'])
            # Save results to csv per agent type
            final_results.to_csv(save_dir+'/testing_variance_results_'+str(agent_type)+'.csv')

        axs[0,0].plot(x,avg_R, color=c, linestyle=l, label=agent_type)
        axs[0,0].fill_between(x,avg_R-avg_R_SE, avg_R+avg_R_SE, color=c, alpha = 0.2)
        axs[0,1].plot(avg_r_mean_sorted,cdf_mean, color=c, linestyle=l)
        axs[1,0].plot(x,cum_R, color=c, linestyle=l)
        axs[1,0].fill_between(x,cum_R-cum_R_SE, cum_R+cum_R_SE, color=c, alpha = 0.2)
        axs[1,1].hist(time_mean, color=c, alpha=0.25)

        axs[0,0].set_xlabel("Episode")
        axs[0,0].set_ylabel('Reward')
        axs[0,0].axes.get_xaxis().set_ticks([0, num_episode])
        axs[0,0].set_title("Mean and Std. Err. of Rolling Avg. R (window size="+str(w)+ " epi)")
        
        axs[0,1].set_ylabel("Cumulative Probability")
        axs[0,1].set_xlabel("Mean Reward per Episode Window")
        axs[0,1].set_title("CDF of Rolling Average R")
        
        axs[1,0].set_xlabel("Episode")
        axs[1,0].set_ylabel('Cumulative Reward')
        axs[1,0].axes.get_xaxis().set_ticks([0, num_episode])
        axs[1,0].set_title("Cumulative R with Std. Err.")
        
        axs[1,1].set_ylabel("Occurence")
        axs[1,1].set_xlabel("Time")
        axs[1,1].set_title("Dist of Time per Episode")

        #ax1.legend(loc=2, bbox_to_anchor=(-0.05, 0), fancybox=True, shadow=True, framealpha=1)
        #ax2.legend(loc=2, bbox_to_anchor=(0, 1), fancybox=True, shadow=True, framealpha=1)
        fig.suptitle("Testing Results for fixed agent: "+str(agent_type)+'\n Variance over '+str(num_repeats)+" runs with randomly generated environment seeds")
        fig.legend(loc='upper right', fancybox=True, shadow=True, framealpha=1)
        fig.set_size_inches(12, 8)
        fig.tight_layout()
        if show_figures == 'Y':
            plt.show(block=False)
            plt.pause(5)
            fig.savefig(save_dir+'/testing_variance_comparison.png', dpi=fig.dpi)
            plt.close()
        else:
            fig.savefig(save_dir+'/testing_variance_comparison.png', dpi=fig.dpi)
            plt.close()



    def convergence_analysis(self, results_table_dir:str, save_dir:str, show_figures:str):
        """Produces visual and tabular analysis of individual training results."""
        training_results = pd.read_csv(results_table_dir+'/results.csv')
        training_results.opponent = training_results.opponent.astype(str)
        
        VisualAnalysis = VisualOutput(training_results, save_dir, show_figures, self.window_size)
        VisualAnalysis.convergence_analysis_graph()


import torch

# ------ Agent Imports -----------------------------------------
# Universal Agents
from elsciRL.agents.agent_abstract import Agent, QLearningAgent
from elsciRL.agents.table_q_agent import TableQLearningAgent
from elsciRL.agents.neural_q_agent import NeuralQLearningAgent

AGENT_TYPES = {
    "Qlearntab": TableQLearningAgent,
    "Neural_Q": NeuralQLearningAgent,
    "Neural_Q_2": NeuralQLearningAgent,
    "Neural_Q_language": NeuralQLearningAgent
}

PLAYER_PARAMS = {
    "Qlearntab": ["alpha", "gamma", "epsilon"],
    "Neural_Q": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "Neural_Q_2": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "Neural_Q_language": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"]
}

from elsciRL.encoders.sentence_transformer_MiniLM_L6v2 import LanguageEncoder

# This is the main run functions for elsciRL to be imported
# Defines the train/test operators and imports all the required agents and experiment functions ready to be used
# The local main.py file defines the [adapters, configs, environment] to be input

# TODO: This should be where the environment is initialized and then episode_loop (or train/test) is run
# -> results then passed down to experiment to produce visual reporting (staticmethod)
# -> instruction following approach then becomes alternative form of this file to be called instead
# -> TODO: This means we have multiple main.py types (e.g. with/without convergence measure) so should create a directory and finalise naming for this

class UnsupervisedSearch():
    def __init__(self,Config:dict, LocalConfig:dict, Environment, 
                 number_exploration_episodes:int = 100, sim_threshold:float = 0.9,
                 observed_states:dict=None, instruction_results:dict=None):
        self.ExperimentConfig = Config
        self.LocalConfig = LocalConfig

        self.env = Environment
        self.setup_info = self.ExperimentConfig.data | self.LocalConfig.data # TODO: configs aren't consistent formatting 
        self.training_setups: dict = {}

        # New instruction learning
        if not observed_states:
            self.observed_states:dict = {}
        else:
            self.observed_states = observed_states   
        if not instruction_results:
            self.instruction_results:dict = {}
        else:
            self.instruction_results = instruction_results
            
        # Unsupervised search parameters
        self.enc = LanguageEncoder()
        self.number_exploration_episodes = number_exploration_episodes
        self.sim_threshold = sim_threshold
        self.cos = torch.nn.CosineSimilarity(dim=0) 


    def search(self, instruction:str='', instr_description:str='', action_cap:int=5, re_search_override:bool=False):
        # Trigger re-search
        if re_search_override:
            self.observed_states:dict = {}
        print(" ")
        print("Instruction: ", instruction)
        
        # Create tensor vector of description
        instruction_vector = self.enc.encode(' '.join(instr_description))
        for n, agent_type in enumerate(self.setup_info['agent_select']):
            # We are adding then overriding some inputs from general configs for experimental setups
            train_setup_info = self.setup_info.copy()
            # Override action cap for shorter term sub-goals for faster learning
            train_setup_info['training_action_cap'] = action_cap 
            # ----- State Adapter Choice
            adapter = train_setup_info["adapter_select"][n]
            # ----- Agent parameters
            agent_parameters = train_setup_info["agent_parameters"][agent_type]
            train_setup_info['agent_type'] = agent_type
            train_setup_info['agent_name'] = str(agent_type) + '_' + str(adapter) + '_' + str(agent_parameters)
            train_setup_info['adapter_select'] = adapter
            # ----- init agent
            player = AGENT_TYPES[agent_type](**agent_parameters)
            train_setup_info['agent'] = player
            # -----
            # Set env function to training
            train_setup_info['train'] = True
            # ---------------------------------SEARCH-----------------------------------------
            # EXPLORE TO FIND LOCATION OF SUB-GOAL
            sub_goal = None
            train_setup_info['training_results'] = False
            if not self.observed_states:
                train_setup_info['observed_states'] = False
            else:
                train_setup_info['observed_states'] = self.observed_states
            train_setup_info['experience_sampling'] = False
            train_setup_info['live_env'] = True  
            # Seen sub_goal before & sim above threshold
            if (instruction in self.instruction_results):
                if (agent_type+'_'+adapter) in self.instruction_results[instruction]['env_code']:
                    if (self.instruction_results[instruction]['env_code'][agent_type+'_'+adapter]['sim_score']>=self.sim_threshold):
                        sub_goal = self.instruction_results[instruction]['env_code'][agent_type+'_'+adapter]['sub_goal']
            
            if not sub_goal:
                # If no description -> no sub-goal (i.e. envs terminal goal position)
                if not instr_description: 
                    sub_goal = None
                # If no sub-goal -> find best match of description from env 
                else:
                    # Run on live env if observed states unknown
                    if not self.observed_states:
                        print("Instruction not known, searching for matching observation in environment...")
                        train_setup_info['sub_goal'] = sub_goal
                        # --- 
                        # Set exploration parameters
                        train_setup_info['number_training_episodes'] = self.number_exploration_episodes # Override 
                        # ---
                        # Explore env with limited episodes
                        live_env = self.env(train_setup_info)
                        explore_results = live_env.episode_loop() 
                        train_setup_info['training_results'] = explore_results
                        train_setup_info['observed_states'] = live_env.elsciRL.observed_states
                        train_setup_info['experience_sampling'] = live_env.elsciRL.experience_sampling
                        
                        # Extract visited states from env
                        self.observed_states = live_env.elsciRL.observed_states.copy()
                        # for obs in env.elsciRL.observed_states:
                        #     if obs not in self.observed_states:
                        #         self.observed_states[obs] = env.elsciRL.observed_states[obs]
                        #observed_states = env.elsciRL.observed_states
                    
                    # Compare to instruction vector                            
                    max_sim = -1
                    # Any states that are above threshold 
                    sub_goal_list = []
                    for obs_state in self.observed_states:
                        str_state = self.observed_states[obs_state]
                        str_state_stacked = ' '.join(str_state)
                        t_state = self.enc.encode(str_state_stacked)
                        # Average sim across each sentence in instruction vs state
                        total_sim = 0
                        for instr_sentence in instruction_vector:
                            for state_sentence in t_state:
                                total_sim+=self.cos(state_sentence, instr_sentence)
                        sim = total_sim.item()/(len(instruction_vector)*len(t_state))
                        if sim > max_sim:
                            max_sim  = sim
                        if sim > self.sim_threshold:
                            sub_goal = obs_state # Sub-Goal code
                            sub_goal_list.append(sub_goal)
                    # OR if none above threshold within (1-threshold%) of max sim
                    if max_sim < 0:
                        sub_goal = None
                        self.observed_states = {}
                    elif max_sim < self.sim_threshold:
                        print("Minimum sim for observed states to match instruction not found, using best match instead.")
                        print("Best match sim value = ", max_sim)
                        for obs_state in self.observed_states:
                            str_state = self.observed_states[obs_state]
                            str_state_stacked = ' '.join(str_state)
                            t_state = self.enc.encode(str_state_stacked)
                            total_sim = 0
                            # Average sim across each sentence in instruction vs state
                            for instr_sentence in instruction_vector:
                                for state_sentence in t_state:
                                    total_sim+=self.cos(state_sentence, instr_sentence) 
                            sim = total_sim.item()/(len(instruction_vector)*len(t_state))
                            if sim > max_sim*(self.sim_threshold):
                                sub_goal = obs_state # Temp Sub-Goal as most known similar
                                sub_goal_list.append(sub_goal)

            # Log matching sub_goal with instruction        
            if instruction not in self.instruction_results:
                self.instruction_results[instruction] = {}    
                self.instruction_results[instruction]['env_code'] = {} 
                self.instruction_results[instruction]['action_cap'] = action_cap
                
            if (agent_type+'_'+adapter) not in self.instruction_results[instruction]['env_code']:
                self.instruction_results[instruction]['env_code'][agent_type+'_'+adapter] = {}
                
            self.instruction_results[instruction]['env_code'][agent_type+'_'+adapter]['sub_goal'] = sub_goal_list
            self.instruction_results[instruction]['env_code'][agent_type+'_'+adapter]['sim_score'] = max_sim
            # --------------------------------------------------------------------------------
        return self.observed_states, self.instruction_results 


import random
import pickle
import numpy as np
import pandas as pd
import logging
from typing import List, Tuple, Dict

import torch
from torch import Tensor

from elsciRL.agents.agent_abstract import QLearningAgent


logger = logging.getLogger()
logger.setLevel(logging.INFO)


class TableQLearningAgent(QLearningAgent):
    def __init__(self, alpha: float, gamma: float, epsilon: float, 
                 epsilon_step:float):
        self.alpha: float = alpha
        self.gamma: float = gamma
        self.epsilon_reset = epsilon
        self.epsilon: float = epsilon
        self.epsilon_step: float = epsilon_step
        self.q_table: dict = {}
        self.q_table_max: dict = {}
        self.total_q: float = 0
        self.q_size: int = 0
        self.q_zeros: int = 0
        self.debugging: bool = False

    def save(self) -> List[dict]:
        return [self.q_table, self.q_table_max]
    
    def load(self, saved_agent:List[dict]=[None,None]):
        # Override init q table with pre saved info
        self.q_table:dict = saved_agent[0]
        self.q_table_max:dict = saved_agent[1]
        if not self.q_table:
            self.q_table = {}
        if not self.q_table_max:
            self.q_table_max = {}

    def exploration_parameter_reset(self):
        self.epsilon = self.epsilon_reset

    def clone(self):
        clone = pickle.loads(pickle.dumps(self))
        clone.epsilon = self.epsilon_reset
        return clone

    # Fixed order of variables
    def policy(self, state: Tensor, legal_actions: list) -> str:
        """Agent's decision making for next action based on current knowledge and policy type"""
        # Initialise the agent_id entry if not seen
        state = state.cpu().numpy().flatten()
        state_tuple = tuple(state)
        # Epsilon greedy action selection
        # - Early exploration
        rng = np.random.rand()
        # - State not seen before)
        if state_tuple not in self.q_table:
            if self.debugging:
                print("\n --- State not seen before - ", state_tuple)
            self.q_table[state_tuple] = {}
            action_code = random.choice(legal_actions)
        # - Epsilon greedy random selection
        elif rng < self.epsilon: 
            action_code = random.choice(legal_actions)
            if self.epsilon > 0:
                self.epsilon = self.epsilon - (self.epsilon*self.epsilon_step) # Added epsilon step reduction to smooth exploration to greedy
                if self.epsilon < 0:
                    self.epsilon = 0
        # - No known actions
        elif not self.q_table[state_tuple]:
            if self.debugging:
                print("\n --- No known actions for state - ", state_tuple)
            action_code = random.choice(legal_actions)
        # - Current state not in q_table_max
        elif not self.q_table_max[state_tuple]:
            if self.debugging:
                print("\n --- State not in q_table_max - ", state_tuple)
            action_code = random.choice(legal_actions)
        # --------------------------------------
        # - If best action has zero value, pick random action from unexplored actions
        elif (list(self.q_table_max[state_tuple].items())[0][1]==0):
            # Start with actions not seen before
            legal_actions_filtered = [action for action in legal_actions if action not in self.q_table[state_tuple]]
            if len(legal_actions_filtered) == 0:
                # Then add actions with zero value
                for action in self.q_table[state_tuple].keys():
                    if self.q_table[state_tuple][action] == 0:
                        legal_actions_filtered.append(action)
            # Sample from this new filtered list of actions
            if len(legal_actions_filtered) > 0:
                if self.debugging:
                    print("\n --- Best known action value is zero, following actions have no knowledge: ")
                    print(legal_actions_filtered)
                action_code = random.choice(legal_actions_filtered)
            else:
                action_code = random.choice(legal_actions)
        # - have to have an action we know this is good -> Moved into new q_table_max check
        elif (list(self.q_table_max[state_tuple].items())[0][1]<0):
            if self.debugging:
                print("\n --- Best known action value less than zero: ")
                print(list(self.q_table_max[state_tuple].items())[0])
            action_code = random.choice(legal_actions)

        # -- Not sure if this is helping learning, 
        # -- logically we dont wan't agent making bad actions but small punishment makes 
        # -- all actions bad before more exploration   
        # --------------------------------------
        # - Epsilon Greedy best action selection 
        else:
            #Now uses q_table_max to lookup best known action
            action_code = list(self.q_table_max[state_tuple].items())[0][0]
            if self.debugging:
                print("\n --- Best known action is - ", action_code)
                print(list(self.q_table_max[state_tuple].items())[0])
            # Extra step in case top choice is invalid
            # -- Somewhat expensive but likely to be rare and a limited number of actions to iterate over
            if action_code not in legal_actions:
                logger.info(" ")
                logger.info("Best known action not a valid move, searching for next best alternative for action - ", action_code)
                state_action_df = {}
                i = 0
                for action,value in self.q_table[state_tuple].items():
                    state_action_df[action] = value
                    i = i+1
                state_action_df_sorted = dict(sorted(state_action_df.items(), key=lambda item: item[1], reverse=True))
                action_found = False
                i=0
                for action in list(state_action_df_sorted.keys()):
                    if action in legal_actions:
                        action_code = action
                        action_found = True
                        break
                    else:
                        i = i+1
                        logger.info("Top " + str(i) + " action not a valid move, searching for next best alternative")
                        continue
                # If no valid moves exists, pick random action
                if action_found == False:
                    action_code = random.choice(legal_actions)
        return action_code

    # We now break agent into a policy choice, action is taken in game_env then next state is used in learning function
    def learn(self, state: Tensor, next_state: Tensor, r_p: float, action_code: str) -> float:
        """Given action is taken, agent learns from outcome (i.e. next state)"""
        # Get q value for current state-action pairs
        # - If state-action pair unseen, add this to q_table with value initialized to 0
        state = state.cpu().numpy().flatten()
        state_tuple = tuple(state)
        next_state = next_state.cpu().numpy().flatten()
        next_state_tuple = tuple(next_state)

        if action_code not in self.q_table[state_tuple]:
            q = 0
            self.q_table[state_tuple][action_code] = 0
            self.q_size += 1
            self.q_zeros += 1
        # - Else simply select known value
        else:
            q = self.q_table[state_tuple][action_code]

        # Extract Q value of next state and apply update to current state-action pair
        if next_state_tuple not in self.q_table:
            q_update = q + self.alpha * (r_p + self.gamma * 0 - q)
        else:
            ns_max_q = max(self.q_table[next_state_tuple].values()) if (list(self.q_table[next_state_tuple].values())) else 0
            q_update = q + self.alpha * (r_p + self.gamma * ns_max_q - q)
        # Update the total Q value tallying (optimizing the results calculation).
        self.q_tally_update(state_tuple, action_code, q_update)
        # Now we replace the state-action lookup with new value and update the total tallying
        self.q_table[state_tuple][action_code] = q_update
        
        # New max q table to only log the best action in each state to save runtime
        if state_tuple not in self.q_table_max:
            self.q_table_max[state_tuple] = {}
            self.q_table_max[state_tuple][action_code] = q_update
        elif q_update > list(self.q_table_max[state_tuple].items())[0][1]:
            self.q_table_max[state_tuple] = {}
            self.q_table_max[state_tuple][action_code] = q_update

        return q_update

    # Merged these into one function to save time
    def q_result(self):
        """Summarizes all known values in Q table."""
        q_size = self.q_size
        total_q = self.total_q
        q_zeros = self.q_zeros
        
        # Ignores state-actions pairs with no knowledge
        if (q_size - q_zeros) != 0:
            mean_q = total_q/(q_size - q_zeros)
        else:
            mean_q = 0

        return total_q, mean_q

    def q_tally_update(self, state_tuple: Tuple[float], action_code: str, q_update: float):
        self.total_q += q_update - self.q_table[state_tuple][action_code]
        self.q_zeros -= int(self.q_table[state_tuple][action_code] == 0 and q_update != 0)
        self.q_zeros += int(self.q_table[state_tuple][action_code] != 0 and q_update == 0)


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import os 


def combined_variance_analysis_graph(results_dir:str='', analysis_type='training', 
                            show_figures:str='N', results_to_show:str='all',
                            experiment_names:list=[]):
    
    if results_dir == '':
        raise ValueError("Save directory not specified.")
    analysis_type = analysis_type.lower() # lowercase analysis type input
    # Get sub-dir for each problem-experiment type
    instruction_folders = [os.path.join(results_dir, instr) for instr in os.listdir(results_dir) if os.path.isdir(os.path.join(results_dir, instr))]
    print(instruction_folders)
    variance_results = {}
    for instr_folder_dir in instruction_folders:
        instr_id = instr_folder_dir.split('/')[-1]
        if instr_id not in variance_results.keys():
            variance_results[instr_id] = {}
        problem_folders = [name for name in os.listdir(instr_folder_dir) if os.path.isdir(os.path.join(instr_folder_dir, name))]
        # Find experiment folders
        # - Capture case where there is only one experiment type
        # and so wont have sub-directory for experiments to search
        for experiment_dir in problem_folders:
            if analysis_type == 'training':
                experiment_name = experiment_dir+'_training'
                file_names = [name for name in os.listdir(instr_folder_dir+'/'+experiment_dir) if name[0:25] == 'training_variance_results']
            elif analysis_type == 'testing':
                experiment_name = experiment_dir+'_testing'
                file_names = [name for name in os.listdir(instr_folder_dir+'/'+experiment_dir) if name[0:24] == 'testing_variance_results']
            else:
                raise ValueError("Analysis type must be either 'training' or 'testing'.")

            if experiment_name not in variance_results[instr_id].keys():
                variance_results[instr_id][experiment_name] = {}

            for file in file_names:
                results = pd.read_csv(instr_folder_dir+'/'+experiment_dir+'/'+file)
                agent = results['agent'].iloc[0].split('__')[0]
                if agent not in variance_results[instr_id][experiment_name].keys():
                    variance_results[instr_id][experiment_name][agent] = {}
                variance_results[instr_id][experiment_name][agent]['results'] = results
                variance_results[instr_id][experiment_name][agent]['num_repeats'] = results['num_repeats'].iloc[0]
    
    cycol = 'brgcmykbrgcmyk'
    line_styles = ['solid','dotted','dashed','dashdot', 'solid','dotted','dashed','dashdot',
                   'dashdot','dashdot','dashdot','dashdot','dashdot','dashdot','dashdot','dashdot']
    col = 0
    if results_to_show == 'simple':
        fig, axs = plt.subplots(1,1)
        for i,instr_id in enumerate(list(variance_results.keys())):
            for n,experiment in enumerate(list(variance_results[instr_id].keys())):
                for ag,agent in enumerate(list(variance_results[instr_id][experiment].keys())):
                    results = variance_results[instr_id][experiment][agent]['results']
                    num_episode = np.max(results['episode'])
                    avg_r_mean_sorted = np.sort(results['avg_R_mean'])
                    cdf_mean = 1. * np.arange(len(avg_r_mean_sorted)) / (len(avg_r_mean_sorted) - 1)

                    # Plot RL Reward results for each approach
                    # 1.1 Summary of total REWARD
                    c = cycol[col]
                    l = line_styles[col]
                    if col <= len(cycol):
                        col+=1
                    else:
                        c = np.random.rand(len(x),3)
                        l = 'solid'                
                    x =  results['episode']
                    avg_R = np.array(results['avg_R_mean'])
                    avg_R_SE = np.array(results['avg_R_se'])
                    cum_R = np.array(results['cum_R_mean'])
                    cum_R_SE = np.array(results['cum_R_se'])
                    time_mean = np.array(results['time_mean'])

                    if 'instr_exp' in str(experiment).lower():
                        label = str(instr_id) + ' - ' + str(agent)
                    else:
                        label = 'No Instr - ' + str(agent)

                    axs.plot(x,avg_R, color=c, linestyle=l, label=label)
                    if len(experiment_names) > 0:
                        axs.set_label(experiment_names[n])
                    axs.fill_between(x,avg_R-avg_R_SE, avg_R+avg_R_SE, color=c, alpha = 0.2)
                
        axs.set_xlabel("Episode")
        axs.set_ylabel('Reward')
        axs.axes.get_xaxis().set_ticks([0, num_episode])
        axs.set_title("Mean Reward per Episode")
        fig.legend(loc='upper left', fancybox=True, shadow=True, 
                   framealpha=1, prop={'size': 14})
        fig.set_size_inches(12, 8)
        
    else:
        fig, axs = plt.subplots(2,2)
        for i,instr_id in enumerate(list(variance_results.keys())):
            for n,experiment in enumerate(list(variance_results[instr_id].keys())):
                for ag,agent in enumerate(list(variance_results[instr_id][experiment].keys())):
                    results = variance_results[instr_id][experiment]['results']
                    num_episode = np.max(results['episode'])
                    avg_r_mean_sorted = np.sort(results['avg_R_mean'])
                    cdf_mean = 1. * np.arange(len(avg_r_mean_sorted)) / (len(avg_r_mean_sorted) - 1)

                    # Plot RL Reward results for each approach
                    # 1.1 Summary of total REWARD
                    c = cycol[col]
                    l = line_styles[col]
                    if col <= len(cycol):
                        col+=1
                    else:
                        c = np.random.rand(len(x),3)
                        l = 'solid'                
                    x =  results['episode']
                    avg_R = np.array(results['avg_R_mean'])
                    avg_R_SE = np.array(results['avg_R_se'])
                    cum_R = np.array(results['cum_R_mean'])
                    cum_R_SE = np.array(results['cum_R_se'])
                    time_mean = np.array(results['time_mean'])

                    if 'instr_exp' in str(experiment).lower():
                        label = str(instr_id) + ' - ' + str(agent)
                    else:
                        label = 'No Instr - ' + str(agent)

                    axs[0,0].plot(x,avg_R, color=c, linestyle=l, label=label)
                    axs[0,0].fill_between(x,avg_R-avg_R_SE, avg_R+avg_R_SE, color=c, alpha = 0.2)
                    axs[0,1].plot(avg_r_mean_sorted,cdf_mean, color=c, linestyle=l)
                    axs[1,0].plot(x,cum_R, color=c, linestyle=l)
                    axs[1,0].fill_between(x,cum_R-cum_R_SE, cum_R+cum_R_SE, color=c, alpha = 0.2)
                    axs[1,1].hist(time_mean, color=c, alpha=0.25)
                    if len(experiment_names) > 0:
                        axs.set_label(experiment_names[n])

        axs[0,0].set_xlabel("Episode")
        axs[0,0].set_ylabel('Reward')
        axs[0,0].axes.get_xaxis().set_ticks([0, num_episode])
        axs[0,0].set_title("Mean and Std. Err. of Rolling Avg. R epi)")
        
        axs[0,1].set_ylabel("Cumulative Probability")
        axs[0,1].set_xlabel("Mean Reward per Episode Window")
        axs[0,1].set_title("CDF of Rolling Average R")
        
        axs[1,0].set_xlabel("Episode")
        axs[1,0].set_ylabel('Cumulative Reward')
        axs[1,0].axes.get_xaxis().set_ticks([0, num_episode])
        axs[1,0].set_title("Cumulative R with Std. Err.")
        
        axs[1,1].set_ylabel("Occurence")
        axs[1,1].set_xlabel("Time")
        axs[1,1].set_title("Dist of Time per Episode")

        #ax1.legend(loc=2, bbox_to_anchor=(-0.05, 0), fancybox=True, shadow=True, framealpha=1)
        #ax2.legend(loc=2, bbox_to_anchor=(0, 1), fancybox=True, shadow=True, framealpha=1)
        fig.suptitle("COMBINED VARIANCE ANALYSIS")
        fig.legend(loc='upper right', fancybox=True, shadow=True, framealpha=1)
        fig.set_size_inches(12, 8)
        fig.tight_layout()

    if show_figures == 'Y':
        fig.savefig(results_dir+'/variance_comparison_'+str(analysis_type)+'.png', dpi=100)
        plt.show(block=False)
        plt.pause(5)
        plt.close()
    else:
        fig.savefig(results_dir+'/variance_comparison_'+str(analysis_type)+'.png', dpi=100)
        plt.show()
        plt.close()




from typing import List
from elsciRL.agents.agent_abstract import Agent, QLearningAgent

class ImportHelper:
    def __init__(self, local_setup_info:dict={}) -> None:
        self.setup_info = local_setup_info

    def agent_info(self, STATE_ADAPTER_TYPES:dict={}):
        agent: Agent | QLearningAgent = self.setup_info['agent']
        agent_type: str = self.setup_info['agent_type']
        agent_name: str = self.setup_info['agent_name']
        if self.setup_info['adapter_select'] in STATE_ADAPTER_TYPES:
            agent_state_adapter = STATE_ADAPTER_TYPES[self.setup_info['adapter_select']](setup_info=self.setup_info)
        else:
            agent_state_adapter = ''
        return agent, agent_type, agent_name, agent_state_adapter

    def parameter_info(self):
        num_train_episodes: int = self.setup_info['number_training_episodes']
        num_test_episodes: int = self.setup_info['number_test_episodes']
        try:
            training_action_cap: int = self.setup_info['training_action_cap']
            testing_action_cap: int = self.setup_info['testing_action_cap']
        except:
            if 'action_limit' in self.setup_info:
                training_action_cap: int = self.setup_info['action_limit']
                testing_action_cap: int = self.setup_info['action_limit']
            elif 'action_cap' in self.setup_info:
                training_action_cap: int = self.setup_info['action_cap']
                testing_action_cap: int = self.setup_info['action_cap']
            else:
                print('No action cap specified, using default values')
                training_action_cap: int = 1000
                testing_action_cap: int = 1000
        reward_signal: List[int] = self.setup_info['reward_signal'] 

        return num_train_episodes, num_test_episodes, training_action_cap, testing_action_cap, reward_signal
    
    def training_flag(self):
        train: bool = self.setup_info['train']
        return train

    def live_env_flag(self):
        live_env: bool = self.setup_info['live_env']
        observed_states: bool = self.setup_info['observed_states']
        experience_sampling: bool = self.setup_info['experience_sampling']
        return live_env, observed_states, experience_sampling



#  Define data through class function so it can be called within package
# Instead of using a .json file which is hard to load from local install
# NOTE: MAKE SURE TO TRUST REPOSITORIES BEFORE RUNNING CODE
# - Can set branch to specific commit to ensure no changes are made without knowledge
#   |-----> changed to commit id which is tied to branch and more stable
# - Compatibility defined to a single engine file
#   |-----> Adapters must be compatible with the given engine
# - Experiment configs are defined in the experiment_configs folder
#   |-----> NOTE: AT LEAST TWO EXPERIMENT CONFIGS MUST BE DEFINED
#       |-----> This is so that it triggers the selection swap on the server side
class Applications:
    def __init__(self):
        self.data ={
            "Sailing":{
                "github_user": "pdfosborne",
                "repository": "elsciRL-App-Sailing",
                "commit_id": "cdf4be9",
                "engine_folder": "environment",
                "engine_filename": "engine",
                "config_folder": "configs",
                "experiment_config_filenames": {"quick-test":"testing.json", 
                                                "Osborne-2024":"config.json"},
                "local_config_filenames": {"easy":"config_local.json"},
                "local_adapter_folder": "adapters",
                "adapter_filenames": {"default":"default", "language":"language"},
                "local_analysis_folder": "analysis",
                "local_analysis_filenames": {"sailing_graphs":"sailing_graphs"},
                "prerender_data_folder": "prerender",
                "prerender_data_filenames": {"observed_states":"observed_states.txt"},
                "prerender_image_filenames": {"Setup":"sailing_setup.png"}
                },
            "Classroom":{
                "github_user": "pdfosborne",
                "repository": "elsciRL-App-Classroom",
                "commit_id": "51b2240",
                "engine_folder": "environment",
                "engine_filename": "engine",
                "config_folder": "configs",
                "experiment_config_filenames": {"default":"config.json"},
                "local_config_filenames": {"classroom_A":"classroom_A.json"},
                "local_adapter_folder": "adapters",
                "adapter_filenames": {"default":"default", "classroom_A_language":"classroom_A_language"},
                "local_analysis_folder": "analysis",
                "local_analysis_filenames": {},
                "prerender_data_folder": "prerender",
                "prerender_data_filenames": {"observed_states":"observed_states.txt"},
                "prerender_image_filenames": {"Classroom_A_Setup":"Classroom_A_Summary.png"}
                },
            "Gym-FrozenLake":{
                "github_user": "pdfosborne",
                "repository": "elsciRL-App-GymFrozenLake",
                "commit_id": "1fd52ea",
                "engine_folder": "environment",
                "engine_filename": "engine",
                "config_folder": "configs",
                "experiment_config_filenames": {"quick_test":"fast_agent.json", "Osborne2024_agent":"Osborne2024_agent.json"},
                "local_config_filenames": {"Osborne2024_env":"Osborne2024_env.json"},
                "local_adapter_folder": "adapters",
                "adapter_filenames": {"numeric_encoder":"numeric", "language":"language"},
                "local_analysis_folder": "analysis",
                "local_analysis_filenames": {},
                "prerender_data_folder": "prerender",
                "prerender_data_filenames": {"observed_states":"observed_states.txt"},
                "prerender_image_filenames": {"FrozenLake_Setup":"FrozenLake_4x4.png"}
                }
        }


# Sailing Simulator
# - https://github.com/topics/sailing-simulator
# - Simple sailing simulator from https://github.com/PPierzc/ai-learns-to-sail
#   - https://github.com/PPierzc/ai-learns-to-sail/blob/master/tasks/channel.py
import io
import numpy as np
import matplotlib.pyplot as plt
from elsciRL.examples.environments.sailing_image import SailingImageData

class Engine:
    """Defines the environment function from the generator engine.
       Expects the following:
        - reset() to reset the env a start position(s)
        - step() to make an action and update the game state
        - legal_moves_generator() to generate the list of legal moves
    """
    def __init__(self, local_setup_info:dict={}) -> None:
        """Initialize Engine"""
        #self.Environment = "Engine Initialization"
        self.x_limit = 10
        self.y_limit = local_setup_info["y_limit"]
        self.angle_limit = np.pi / 2
        self.supervised_rewards = local_setup_info["supervised_rewards"]
        # Precision parameter
        self.obs_precision = local_setup_info["obs_precision"]

        # Ledger of the environment with meta information for the problem
        ledger_required = {
            'id': 'Unique Problem ID',
            'type': 'Language/Numeric',
            'description': 'Problem Description',
            'goal': 'Goal Description'
            }
        
        ledger_optional = {
            'reward': 'Reward Description',
            'punishment': 'Punishment Description (if any)',
            'state': 'State Description',
            'constraints': 'Constraints Description',
            'action': 'Action Description',
            'author': 'Author',
            'year': 'Year',
            'render_data':{'render_mode':'rgb_array', 
                           'render_fps':4}
        }
        ledger_gym_compatibility = {
            # Limited to discrete actions for now, set to arbitrary large number if uncertain
            'action_space_size':2, 
        }
        self.ledger = ledger_required | ledger_optional | ledger_gym_compatibility
        # Initialize history
        self.action_history = []
        self.obs_history = []
        
        

    # --------------------------
    # Defined functions used by engine source
    @staticmethod
    def vel(theta, theta_0=0, theta_dead=np.pi / 12):
        return 1 - np.exp(-(theta - theta_0) ** 2 / theta_dead)
    
    @staticmethod
    def rew(theta, theta_0=0, theta_dead=np.pi / 12):
        return Engine.vel(theta, theta_0, theta_dead) * np.cos(theta)
    # --------------------------

    def reset(self, start_obs:str=None, render_dir:str=None):
        """Fully reset the environment."""
        # Allow reset to be at fixed start position or random
        if start_obs:
            self.x = np.round(float(start_obs.split('_')[0]),self.obs_precision)
            self.angle = np.round(float(start_obs.split('_')[1]),1)
        else:
            self.x = 0 #np.round(np.random.randint(-9.9, 9.9),4) # Changed to rand_int to reduce num of start states
            self.angle = 0  # always start with angle 0
        self.y = 0
        obs = "{n:.{d}f}".format(n=self.x, d=self.obs_precision)+'_'+"{:0.1f}".format(self.angle)
        
        if render_dir:
                
            # SHOW PRETTY IMAGE OF PROBLEM
            raw_image = SailingImageData['data'].split(",")

            width = 240
            height = 300

            full_array = []
            column_counter = 0
            row = []
            pixel_counter = 0
            pixel_list = []
            for input_item in raw_image:
                pixel_item = int(input_item.replace(" ",""))
                if pixel_counter == 3:
                    # new pixel and reset pixel counter
                    pixel_counter = 0
                    pixel_list = []
                    # Add 3-d pixel to row
                    if column_counter == width:
                        # Add row to full array
                        full_array.append(row)
                        # new row and reset column counter
                        column_counter = 0
                        row = []
                
                    row.append(pixel_list)
                    column_counter+=1
                
                pixel_list.append(pixel_item)
                pixel_counter+=1

            render = np.array(full_array)
            plt.imshow(render, interpolation='nearest')
            plt.axis('off')
            plt.title("Sailing Simulation \n Simple River with Fixed Wind Direction")
            plt.show()
            plt.pause(5)
            plt.savefig(render_dir,bbox_inches='tight')
            plt.close()
        
        return obs

    
    def step(self, state:any=None, action:any=None):
        """Enact an action."""
        self.action_history.append(action)
        a = [-0.1, 0.1][action]
        # Observation space
        self.x += np.round((Engine.vel(self.angle + a) * np.sin(self.angle + a)),self.obs_precision) # Round x to Ndp
        self.y += np.round((Engine.vel(self.angle + a) * np.cos(self.angle + a)),4) # Round y to 4dp
        self.angle = np.round(self.angle+a,1) 
        #obs = str(self.x)+'_'+str(self.angle)
        obs = "{n:.{d}f}".format(n=self.x, d=self.obs_precision)+'_'+"{:0.1f}".format(self.angle) # fix - https://docs.python.org/3.4/library/string.html#format-specification-mini-language
        self.obs_history.append(obs)
        # Reward signal
        # - Added flag for whether we give agent immediate positive reward
        # - Update: Added scale factor if using supervised rewards to not override goal rewards
        if self.supervised_rewards=="True":
            reward = Engine.rew(self.angle)/10
        else:
            reward = 0

        # Termination signal
        # - Source: Terminal only on hitting piers/walls, otherwise continues to action limit
        # - Update: Add terminal state if y > 25 (or another arbitrary value)
        # - Update: Limit angle to [-90,90] degrees (i.e. no backwards sailing)
        if np.abs(self.x)>self.x_limit:
            reward = -1
            terminated = True
        elif np.abs(self.y)>self.y_limit:
            reward = 1
            terminated = True
        elif np.abs(self.y)<0:
            reward = -1
            terminated = True
        elif np.abs(self.angle)>self.angle_limit:
            #print("\n \t - Angle limit reached")
            reward = -1
            terminated = True
        else:
            terminated = False
        
        return obs, reward, terminated, {}

    def legal_move_generator(self, obs:any=None):
        """Define legal moves at each position"""
        # Action space: [0,1] for turn slightly left or right
        # - Kept as binary but might be better as continuous [-0.1, 0.1]
        legal_moves = [0, 1]
        return legal_moves
    
    def render(self, state:any=None):
        """Render the environment."""            
        #render = print("Current State: ", state, " | Action History: ", self.action_history)
        # state = x_angle
        x = self.x
        y = self.y
        angle = self.angle
        # Angle is bearing into wind -pi/2 < angle < pi/2
        if angle < np.pi/2:
            U = np.sin(angle)
            V = np.cos(angle)
        elif angle == np.pi/2:
            U = 1
            V = 0
        elif angle == -np.pi/2:
            U = -1
            V = 0
        else:
            U = np.sin(angle)
            V = -np.cos(angle)

        DPI = 128
        fig, ax = plt.subplots(figsize=(5,5), dpi = DPI)
        ax.scatter(x,y,c='b',marker='x',alpha=1)
        ax.quiver(x,y,U,V,angles='uv',scale_units='xy')
        if y > 1:
            ax.text(x+0.5,y-1,'Sailboat',color='b')

        # Draw wind direction
        ax.quiver(0,25,0,-1,angles='uv',scale_units='xy',color='r')
        ax.text(0,25.25,'Wind',color='r')


        ax.plot([10,10],[0,25],'r')
        ax.plot([-10,-10],[0,25],'r')
        ax.set_title("Sailboat Position with Direction against Wind")
        ax.set_xlabel("Horizontal Position (x)")
        ax.set_ylabel("Vertical Position (y)")
        # Save as rgba array 
        # https://stackoverflow.com/questions/7821518/save-plot-to-numpy-array
        

        fig.canvas.draw()
        # data = np.frombuffer(fig.canvas.tostring_argb(), dtype=np.uint8)
        # render = data.reshape(fig.canvas.get_width_height()[::-1] + (4,))

        buf = fig.canvas.buffer_rgba()
        data = np.asarray(buf)
        render = data.reshape(fig.canvas.get_width_height()[::-1] + (4,))
        return render

    def close(self):
        """Close/Exit the environment."""
        print("Environment Closed")


# Get application data
from elsciRL.application_suite.import_data import Applications
from elsciRL.application_suite.import_tool import PullApplications

# Get search method
from elsciRL.GUI.elsciRL_demo_search import elsciRLSearch as elsci_search
import os
import json
from datetime import datetime

class Prerender:
    def __init__(self) -> None:
         # Get application data
        imports = Applications().data
        self.possible_applications = list(imports.keys())
        pull_data = PullApplications()
        self.pull_app_data = pull_data.pull(problem_selection=self.possible_applications)
        self.ExperimentConfig = pull_data.setup()

        

    def get_observed_states(self, engine, local_config:dict={}, 
                            adapters:dict={}, selected_adapter:list=[''], 
                            num_explor_episodes:int=1000):
            
            # Search Save Directory
            if not os.path.exists('./prerender-data'):
                os.mkdir('./prerender-data')
                            
            time = datetime.now().strftime("%d-%m-%Y_%H-%M")
            save_dir = './prerender-data/' + str('search') + '_' + time
            if not os.path.exists(save_dir):                
                os.mkdir(save_dir)
            # UPDATE EXPERIMENT CONFIG FOR SEARCH
            # - only use q learn tab agent
            selected_agents =  ['Qlearntab']
            self.ExperimentConfig.update({
                'number_training_episodes': int(num_explor_episodes),
                'agent_select': selected_agents,
                'adapter_select': selected_adapter             
            })

            elsci_run = elsci_search(Config=self.ExperimentConfig,
                                        LocalConfig=local_config,
                                        Engine=engine, Adapters=adapters,
                                        save_dir=save_dir,
                                        number_exploration_episodes=num_explor_episodes,
                                        match_sim_threshold=0.9,
                                        observed_states=None)
            observed_states = elsci_run.search()
            print(f"\nNumber of observed states: {len(observed_states)}")
            with open(os.path.join(save_dir, 'observed_states.txt'), 'w') as f:
                json.dump(observed_states, f)


    def run(self):

        # Allow terminal input to select application
        print("Select an application from the following options:")
        for i, app in enumerate(self.possible_applications):
            print(f"{i + 1}: {app}")

        selected_index = int(input("Enter the number of the application you want to select: ")) - 1
        selected_application = self.possible_applications[selected_index]

        # Allow terminal input to select configuration
        print("Select a configuration from the following options:")
        for i, config in enumerate(self.pull_app_data[selected_application]['local_configs']):
            print(f"{i + 1}: {config}")

        config_input_index = int(input("Enter the number of the configuration you want to select: ")) - 1
        config_input = list(self.pull_app_data[selected_application]['local_configs'].keys())[config_input_index]

        # Allow terminal input to select adapter
        adapter_list = self.pull_app_data[selected_application]['adapters']
        print("Select an adapter from the following options:")
        for i, adapter in enumerate(adapter_list):
            print(f"{i + 1}: {adapter}")

        selected_adapter_index = int(input("Enter the number of the adapter you want to select: ")) - 1
        selected_adapter = [list(adapter_list.keys())[selected_adapter_index]]

        # Get data for the selected application
        engine = self.pull_app_data[selected_application]['engine']
        local_config = self.pull_app_data[selected_application]['local_configs'][config_input]
        adapters = self.pull_app_data[selected_application]['adapters']
        
        num_explor_episodes = int(input("Enter the number of the exploration episodes: "))

        print("--------------------------------")
        print("-Selected options-")
        print(f"-- Selected application: {selected_application}")
        print(f"-- Selected configuration: {config_input}")
        print(f"-- Selected adapter: {selected_adapter[0]}")
        print(f"-- Number of exploration episodes: {num_explor_episodes}")
        print("--------------------------------")

        self.get_observed_states(engine, local_config, 
                                adapters, selected_adapter,
                                num_explor_episodes)


import os
from typing import List
import json
import torch
import time

import matplotlib.pyplot as plt
# ------ Interaction Protocol -----------------------------------
from elsciRL.interaction_loops.standard import StandardInteractionLoop
# ------ Experiment Import --------------------------------------
from elsciRL.evaluation.standard_report import Evaluation
# ------ Evaluation Metrics -----------------------------------------
from elsciRL.analysis.convergence_measure import Convergence_Measure
# ------ Instruction Inputs -----------------------------------------
from elsciRL.instruction_following.elsciRL_instr_input import elsciRLInput as TerminalInput
# ------ Agent Imports -----------------------------------------
# Universal Agents
from elsciRL.agents.agent_abstract import Agent, QLearningAgent
from elsciRL.agents.table_q_agent import TableQLearningAgent
from elsciRL.agents.neural_q_agent import NeuralQLearningAgent
from elsciRL.agents.random_agent import RandomAgent

# TODO: Enable any number of the same agent types with varying parameters
AGENT_TYPES = {
    "Qlearntab": TableQLearningAgent,
    "Neural_Q": NeuralQLearningAgent,
    "Neural_Q_2": NeuralQLearningAgent,
    "Neural_Q_language": NeuralQLearningAgent,
    "Random": RandomAgent
}

PLAYER_PARAMS = {
    "Qlearntab": ["alpha", "gamma", "epsilon"],
    "Neural_Q": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "Neural_Q_2": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "Neural_Q_language": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "Random":[]
}

# TODO use encoder defined by config not manual import
from elsciRL.encoders.sentence_transformer_MiniLM_L6v2 import LanguageEncoder

# This is the main run functions for elsciRL to be imported
# Defines the train/test operators and imports all the required agents and experiment functions ready to be used
# The local main.py file defines the [adapters, configs, environment] to be input

# This should be where the environment is initialized and then episode_loop (or train/test) is run
# -> results then passed down to experiment to produce visual reporting (staticmethod)
# -> instruction following approach then becomes alternative form of this file to be called instead

class elsciRLSearch:
    def __init__(self, Config:dict, LocalConfig:dict, 
                 Engine, Adapters:dict,
                 save_dir:str, 
                 num_plans:int=1, number_exploration_episodes:int = 100, sim_threshold:float = 0.9,
                 feedback_increment=0.5, feedback_repeats:int=1,
                 observed_states:dict=None, instruction_results:dict=None,
                 instruction_input_interface:str='terminal'):
        self.ExperimentConfig = Config
        self.LocalConfig = LocalConfig

        self.engine = Engine
        self.adapters = Adapters
        self.env = StandardInteractionLoop
        self.setup_info:dict = self.ExperimentConfig['data'] | self.LocalConfig['data']  
        self.training_setups: dict = {}

        save_dir_extra = save_dir.split("/")[-1]
        save_dir = '/'.join(save_dir.split("/")[:-1])
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)
        self.save_dir = save_dir+'/'+save_dir_extra
        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir)

        # New instruction learning
        if not observed_states:
            self.observed_states:dict = {}
        else:
            self.observed_states = observed_states   
        if not instruction_results:
            self.instruction_results:dict = {}
        else:
            self.instruction_results = instruction_results
            
        # Unsupervised search parameters
        self.enc = LanguageEncoder()
        self.number_exploration_episodes: int = number_exploration_episodes
        self.sim_threshold: float = sim_threshold
        self.cos = torch.nn.CosineSimilarity(dim=0)
        # elsciRL inits
        self.num_plans = num_plans
        #self.feedback_layer_form: tensor = torch.zeros(self.enc.encode(['']).size()) # Init base sizing of tensor produced by language encoder        
        self.feedback_increment: float = feedback_increment
        self.feedback_results: dict = {}
        self.feedback_repeats: int = feedback_repeats
        # Instruction input
        self.instruction_input_interface = instruction_input_interface
        if self.instruction_input_interface.lower() == 'terminal':
            self.elsciRL_input = TerminalInput()
        else:
            print("NOTE: Instruction input not defined, default set to terminal.")
            self.elsciRL_input = TerminalInput()


    def search(self, action_cap:int=5, re_search_override:bool=False, simulated_instr_goal:any=None):
        device = "cuda" if torch.cuda.is_available() else "cpu" 
        # Trigger re-search
        if re_search_override:
            self.observed_states:dict = {}
        
        #instruction_vector = torch.rand(observed_states[list(observed_states.keys())[0]].size()) # NEED TO FIND A METHOD TO VECTORIZE OBSERVED STATES AND INSTRUCTIONS TO COMPARE
        # - TODO: Only use first given agent to do exploration - repeating search for multiple agents doesn't make sense
        for n, agent_type in enumerate(self.setup_info['agent_select']): 
            # We are adding then overriding some inputs from general configs for experimental setups
            train_setup_info = self.setup_info.copy()
            # Override action cap for shorter term sub-goals for faster learning
            train_setup_info['training_action_cap'] = action_cap 
            # ----- State Adapter Choice
            adapter = train_setup_info["adapter_select"][n]
            # ----- Agent parameters
            agent_parameters = train_setup_info["agent_parameters"][agent_type]
            train_setup_info['agent_type'] = agent_type
            train_setup_info['agent_name'] = str(agent_type) + '_' + str(adapter) + '_' + str(agent_parameters)
            train_setup_info['adapter_select'] = adapter
            # ----- init agent
            player = AGENT_TYPES[agent_type](**agent_parameters)
            train_setup_info['agent'] = player
            # -----
            # Set env function to training# Repeat training
            train_setup_info['train'] = True
            # --- 
            # Set exploration parameters
            train_setup_info['number_training_episodes'] = self.number_exploration_episodes # Override
            
            # ---------------------------------elsciRL-----------------------------------------
            # Train on Live system for limited number of total episodes
            train_setup_info['training_results'] = False
            if not self.observed_states:
                train_setup_info['observed_states'] = False
            else:
                train_setup_info['observed_states'] = self.observed_states
            train_setup_info['experience_sampling'] = False
            train_setup_info['live_env'] = True 
            # ---------------------------
            # Init environment to define current position
            for p in range(0, self.num_plans):
                print("---")
                print("Plan Number = ", p)
                sample_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
                print("Environment Init Position: ", sample_env.start_obs)
                # New: user input here
                instructions, instr_descriptions = self.elsciRL_input.user_input()
                
                # ---------------------------
                feedback_count = 0
                for repeat in range(0,self.feedback_repeats): # Arbitrary repeat to further reinforce matching state to instr
                    print("===")
                    print("REINFORCEMENT Repeated search num ", repeat+1)
                    for i,instr in enumerate(instructions):
                        
                        if i == 0:
                            instruction = str(sample_env.start_obs).split(".")[0] + "---" + instr
                        else:
                            instruction = instructions[i-1] + "---" + instr
                        
                        instr_description = instr_descriptions[i]
                        
                        if type(instr_description) == type(''):
                            instr_description = instr_description.split('.')
                            instr_description = list(filter(None, instr_description))
                        # Create tensor vector of description
                        instruction_vector = self.enc.encode(instr_description)
                        print("Size of Encoded Instruction Tensor")
                        print(instruction_vector.size())
                        # EXPLORE TO FIND LOCATION OF SUB-GOAL
                        sub_goal = None
                        # ---------------------------
                        # Overrides on prior knowledge
                        # Seen sub_goal before & sim above threshold
                        if (instruction in self.instruction_results):
                            if (agent_type+'_'+adapter) in self.instruction_results[instruction]:
                                # We use feedback layer even if sub_goal not a good match
                                feedback_layer = self.instruction_results[instruction][agent_type+'_'+adapter]['feedback_layer']
                                if (self.instruction_results[instruction][agent_type+'_'+adapter]['sim_score']>=self.sim_threshold):
                                    sub_goal = self.instruction_results[instruction][agent_type+'_'+adapter]['sub_goal'][0]
                                    sub_goal_list = self.instruction_results[instruction][agent_type+'_'+adapter]['sub_goal']
                                    sim = self.instruction_results[instruction][agent_type+'_'+adapter]['sim_score']
                            else:
                                self.instruction_results[instruction][agent_type+'_'+adapter] = {}
                                feedback_layer = torch.zeros(instruction_vector.size()).to(device)
                            self.instruction_results[instruction][agent_type+'_'+adapter]['count'] = self.instruction_results[instruction][agent_type+'_'+adapter]['count']+1
                        else:
                            self.instruction_results[instruction] = {}    
                            self.instruction_results[instruction][agent_type+'_'+adapter] = {} 
                            self.instruction_results[instruction][agent_type+'_'+adapter]['count'] = 1
                            self.instruction_results[instruction][agent_type+'_'+adapter]['action_cap'] = action_cap
                            feedback_layer = torch.zeros(instruction_vector.size()).to(device)
                            self.instruction_results[instruction][agent_type+'_'+adapter]['feedback_layer'] = feedback_layer
                        # ---------------------------
                        search_count=0
                        while not sub_goal:
                            sim_delta = 1 # Parameter that stops updates when change to feedback is small
                            # If no description -> no sub-goal (i.e. envs terminal goal position)
                            if not instr_description: 
                                sub_goal = None
                                # If no sub-goal -> find best match of description from env 
                            else:
                                search_count+=1
                                print("------") 
                                print("Search: ", search_count)
                                
                                # Only run on live env if observed states empty
                                if not self.observed_states:
                                    train_setup_info['sub_goal'] = sub_goal # None
                                    # ---
                                    # Explore env with limited episodes
                                    # Environment now init here and called directly in experimental setup loop
                                    # - setup elsciRL info
                                    # Train on Live system for limited number of total episodes
                                    train_setup_info['live_env'] = True                        
                                    live_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
                                    explore_results = live_env.episode_loop()
                                    train_setup_info['training_results'] = explore_results
                                    train_setup_info['observed_states'] = live_env.elsciRL.observed_states
                                    train_setup_info['experience_sampling'] = live_env.elsciRL.experience_sampling
                                    # Extract visited states from env
                                    self.observed_states = live_env.elsciRL.observed_states

                                # Compare to instruction vector                            
                                max_sim = -1
                                # all states that are above threshold 
                                sub_goal_list = []
                                for obs_state in self.observed_states:
                                    str_state = self.observed_states[obs_state]
                                    #str_state_stacked = ' '.join(str_state)
                                    t_state = self.enc.encode(str_state)
                                    # ---
                                    total_sim = 0
                                    dim_count = 0 # For some reason encoder here is adding extra dimension                                        
                                    for idx,instr_sentence in enumerate(instruction_vector):
                                        feedback_layer_sent = feedback_layer[idx]
                                        for state_sentence in t_state:
                                            total_sim+=self.cos(torch.add(state_sentence, feedback_layer_sent), instr_sentence)
                                            dim_count+=1
                                    #print("Observation: ", str_state)
                                    #print("Tensor check: ", t_state.size(), " --- ", instruction_vector.size(), " --- ", feedback_layer.size())
                                    #print("Sim result: ", total_sim.item(), '---', dim_count)
                                    
                                    sim = 0 if dim_count==0 else total_sim.item()/dim_count
                                    if sim > max_sim:
                                        max_sim  = sim
                                        sub_goal_max = obs_state
                                        sub_goal_max_t = t_state
                                    if sim >= self.sim_threshold:
                                        sub_goal = obs_state # Sub-Goal code
                                        sub_goal_t = t_state
                                        sub_goal_list.append(sub_goal)
                
                                # OR if none above threshold matching max sim
                                if max_sim < self.sim_threshold:
                                    sub_goal = sub_goal_max
                                    sub_goal_t = sub_goal_max_t                                    
                                    # Find all states that have same sim as max
                                    for obs_state in self.observed_states:
                                        str_state = self.observed_states[obs_state]
                                        #str_state_stacked = ' '.join(str_state)
                                        t_state = self.enc.encode(str_state)
                                        # ---
                                        total_sim = 0
                                        # Average sim across each sentence in instruction vs state
                                        dim_count = 0
                                        for idx,instr_sentence in enumerate(instruction_vector):
                                            feedback_layer_sent = feedback_layer[idx]
                                            for state_sentence in t_state:
                                                total_sim+=self.cos(torch.add(state_sentence, feedback_layer_sent), instr_sentence)
                                                dim_count+=1
                                        sim = 0 if dim_count==0 else total_sim.item()/dim_count
                                        if sim >= (max_sim):
                                            sub_goal_list.append(obs_state)

                                if max_sim < self.sim_threshold:
                                    print("Minimum sim for observed states to match instruction not found, using best match instead. Best match sim value = ", max_sim )
                            # If adapter is poor to match to instruction vector none of them observed states match
                            if (max_sim<-1)|(sub_goal is None):#|(max_sim>1):
                                print("All observed states result in similarity outside bounds (i.e. strongly opposite vectors to instruction). Re-starting Search.")
                                print("Max sim found = ", max_sim)
                                sub_goal = None
                                self.observed_states = {}
                            elif (sim_delta<0.05)|(search_count>=3):
                                print("Change in sim less than or equal to delta cap or search cap reached, assume goal-state not observed. Re-starting Search.")
                                if simulated_instr_goal:
                                    print("-- Known sub_goal position: ", simulated_instr_goal)
                                print("-- Best match: ", sub_goal_max)
                                sub_goal = None
                                self.observed_states = {}
                                search_count = 0
                            else:
                                if not simulated_instr_goal:
                                    
                                    for s_g in sub_goal_list:
                                        print(self.observed_states[s_g], ": ", s_g)
                                    print(" ")
                                    print("===========")
                                    print("Instruction: ", instr)
                                    print(sub_goal)
                                    print("Best match state for instruction \n ______ \n Adapted form: \n\t - ", self.observed_states[sub_goal], " \n Engine observation: \n\t - " , sub_goal)
                                    print("Full list of matching states...",)
                                    print("")
                                    print("Tensor check: Size of Instruction vs Best match: ", instruction_vector.size(), " --- ", sub_goal_t.size())
                                    print("")
                                    feedback = input("-- Does this match the expectation instruction outcome? (Y/N)")
                                    feedback_count+=1 
                                else:
                                    if type(simulated_instr_goal[0]) != type(sub_goal_max):
                                        print("- ERROR: Typing of simulated sub-goal check does not match typing of state from environment, please correct this.")
                                        print(type(simulated_instr_goal[0])," - ", type(sub_goal_max))
                                        #print("Observed States Examples: ", self.observed_states(list(self.observed_states.keys())[0:5]))
                                        exit()
                                    else:
                                        if sub_goal_max in simulated_instr_goal:
                                            print("- Simulated sub-goal match found.")
                                            print("-- Known sub_goal position: ", simulated_instr_goal)
                                            print("-- Unsupervised state best match: ", sub_goal_max)
                                            feedback = 'Y'
                                        else:
                                            print("- Match NOT found.")
                                            print("-- Known sub_goal position: ", simulated_instr_goal)
                                            print("-- Unsupervised state best match: ", sub_goal_max)
                                            feedback = 'N'
                                
                                if (feedback.lower() == 'y')|(feedback.lower() == 'yes'):
                                    for idx,instr_sentence in enumerate(instruction_vector):
                                        feedback_layer_sent = feedback_layer[idx]
                                        for sentence in sub_goal_t:
                                            feedback_layer[idx] = torch.add(feedback_layer_sent, self.feedback_increment*(torch.sub(instr_sentence, sentence))) 
                                    total_sim = 0
                                    # Average sim across each sentence in instruction vs state
                                    dim_count = 0
                                    for idx,instr_sentence in enumerate(instruction_vector):
                                        feedback_layer_sent = feedback_layer[idx]
                                        for state_sentence in t_state:
                                            total_sim+=self.cos(torch.add(state_sentence, feedback_layer_sent), instr_sentence)
                                            dim_count+=1
                                    sim = 0 if dim_count==0 else total_sim.item()/dim_count
                                    sim_delta = sim-max_sim
                                    print("--- Change in sim results with POSITIVE reinforcement of correct state match =", sim_delta)
                                    print("--- New Sim Value = ",  sim)
                                else:
                                    for idx,instr_sentence in enumerate(instruction_vector):
                                        feedback_layer_sent = feedback_layer[idx]
                                        for sentence in sub_goal_t:
                                            feedback_layer[idx] = torch.sub(feedback_layer_sent, self.feedback_increment*(torch.sub(instr_sentence, sentence)))
                                    total_sim = 0
                                    # Average sim across each sentence in instruction vs state
                                    dim_count = 0
                                    for idx,instr_sentence in enumerate(instruction_vector):
                                        feedback_layer_sent = feedback_layer[idx]
                                        for state_sentence in t_state:
                                            total_sim+=self.cos(torch.add(state_sentence, feedback_layer_sent), instr_sentence)
                                            dim_count+=1
                                    sim = 0 if dim_count==0 else total_sim.item()/dim_count
                                    sim_delta = sim-max_sim
                                    print("--- Change in sim results with NEGATIVE reinforcement for NO MATCH =", sim_delta)
                                    sub_goal = None
                        if (agent_type+'_'+adapter) not in self.instruction_results[instruction]:
                            self.instruction_results[instruction][agent_type+'_'+adapter] = {}   
                        self.instruction_results[instruction][agent_type+'_'+adapter]['sub_goal'] = sub_goal_list
                        self.instruction_results[instruction][agent_type+'_'+adapter]['sim_score'] = sim
                        self.instruction_results[instruction][agent_type+'_'+adapter]['feedback_layer'] = feedback_layer
                # log results of feedback loop
                if (agent_type+'_'+adapter) not in self.feedback_results:
                    self.feedback_results[agent_type+'_'+adapter] = {}
            
                if p not in self.feedback_results[(agent_type+'_'+adapter)]:
                    self.feedback_results[(agent_type+'_'+adapter)][p] = {}
                    self.feedback_results[(agent_type+'_'+adapter)][p]['instr_count'] = i+1
                    self.feedback_results[(agent_type+'_'+adapter)][p]['feedback_count'] = feedback_count
                else:
                    self.feedback_results[(agent_type+'_'+adapter)][p]['feedback_count'] = feedback_count + self.feedback_results[(agent_type+'_'+adapter)][p]['feedback_count']
        # --------------------------------------------------------------------------------
        # Quick Visual Analysis of Feedback results
        goal = str(sample_env.start_obs).split(".")[0] + '_GOAL'
         
        plan_numbers = []
        feedback_required = []
        for plan_num in self.feedback_results[agent_type+'_'+adapter]:
            plan_numbers.append(int(plan_num))
            instr_count = self.feedback_results[agent_type+'_'+adapter][plan_num]['instr_count']
            feedback_count = self.feedback_results[agent_type+'_'+adapter][plan_num]['feedback_count']
            feedback_avg = feedback_count/instr_count
            feedback_required.append(feedback_avg)
        plt.plot(plan_numbers,feedback_required, label='Num Searches')
        plt.title('Amount of Feedback for each Instruction by Plan')
        plt.ylabel('Feedback Needed per Instr')
        plt.xlabel('Plan Number')
        plt.tight_layout()
        plt.savefig(self.save_dir+'/reinforcement_results.png', dpi=100)
        plt.show()
        plt.close()
        
        
        # Save instruction inputs
        instruction_predictions = {}
        for instr in self.instruction_results:
            
            if instr not in instruction_predictions:
                instruction_predictions[instr] = {}
            for n, agent_type in enumerate(self.setup_info['agent_select']):
                adapter = self.setup_info["adapter_select"][n]
                
                if (agent_type+'_'+adapter) not in instruction_predictions:
                    instruction_predictions[instr][(agent_type+'_'+adapter)] = {}

                instruction_predictions[instr][(agent_type+'_'+adapter)]['count'] = self.instruction_results[instr][(agent_type+'_'+adapter)]['count']
                instruction_predictions[instr][(agent_type+'_'+adapter)]['sim_score'] = self.instruction_results[instr][(agent_type+'_'+adapter)]['sim_score']
                instruction_predictions[instr][(agent_type+'_'+adapter)]['sub_goal'] = self.instruction_results[instr][(agent_type+'_'+adapter)]['sub_goal']



        with open(self.save_dir+'/instruction_predictions.json', 'w') as f:
            json.dump(instruction_predictions, f)
        
        print("-- Complete Instruction Results --")
        print(self.instruction_results)

        return self.observed_states, self.instruction_results


from typing import Dict, List
import pandas as pd
import torch
from torch import Tensor
# StateAdapter includes static methods for adapters
from elsciRL.encoders.sentence_transformer_MiniLM_L6v2 import LanguageEncoder

class LanguageAdapter:
    _cached_state_idx: Dict[str, int] = dict()

    def __init__(self, setup_info:dict={}):
        # Language encoder doesn't require any preset knowledge of env to use
        self.encoder = LanguageEncoder()
        self.obs_mapping = {0:'You are at the start position.', 1:'You are on ice.', 2:'You are on ice.', 3:'You are on ice.',
                       4:'You are on ice.', 5:'You fell through a hole in the ice!', 6:'You are on ice.', 7:'You fell through a hole in the ice!',
                       8:'You are on ice.', 9:'You are on ice.', 10:'You are on ice.', 11:'You fell through a hole in the ice!',
                       12:'You fell through a hole in the ice!', 13:'You are on ice.', 14:'You are on ice.', 15:'You found the chest!'}
        self.key_found = False
    
    def adapter(self, state:any, legal_moves:list = None, episode_action_history:list = None, encode:bool = True, indexed: bool = False) -> Tensor:
        """ Use Language name for every piece name for current board position """
        # ---
        # Convert to lanugage
        state = self.obs_mapping[state]
        # ---
        
        # Encode to Tensor for agents
        if encode:
            state_encoded = self.encoder.encode(state=state)
        else:
            state_encoded = state

        if (indexed):
            state_indexed = list()
            for sent in state:
                if (sent not in LanguageAdapter._cached_state_idx):
                    LanguageAdapter._cached_state_idx[sent] = len(LanguageAdapter._cached_state_idx)
                state_indexed.append(LanguageAdapter._cached_state_idx[sent])

            state_encoded = torch.tensor(state_indexed)

        return state_encoded


import os
import logging
import json

class Config:
    def __init__(self, config_file_path: str):
        if (config_file_path):
            with open(config_file_path) as config_file:
                self.data = json.load(config_file)
                self.config_file_path = config_file_path
                
        else:
            self.data = dict()
            self.config_path = ""
            logging.info("No arguments given, using default configuration...")
        
    def __getitem__(self, key: str):
        item = None

        if (key in self.__dict__):
            item = self.__dict__[key]
        else:
            item = self.data[key]

        return item

class ExperimentConfig(Config):
    def __init__(self, config_path: str):
        super(ExperimentConfig, self).__init__(config_path)
        
        # Name setup
        self.name = self.data.get(
            "name", 
            os.path.split(self.config_file_path)[-1].replace(".json", "")
            )
        # Define Problem Type Choice
        self.problem_type = self.data.get("problem_type", "")
        # Specify local config choices to select agents of interest
        self.agent_select = self.data.get("agent_select", ["Qlearntab"])
        
        # ---> We then parse these three inputs to obtain the local config setup info
        # ---> Ideally input is a dict input: setups = { 'Setup1':{"Adapter":"Engine", "Encoder":"Yes", "Agent":"TabQ"},... }

        # Training repeated 
        self.experience_sample_batch_size = self.data.get(
            "experience_sample_batch_ratio", 0)

        # Training repeated 
        self.num_training_episodes = self.data.get("num_training_episodes", 1000)
        self.number_training_repeats = self.data.get("number_training_repeats", 5)

        # Testing repeated
        self.number_test_episodes = self.data.get("number_test_episodes", 100)
        self.number_test_repeats = self.data.get("number_test_repeats", 5)
        self.test_agent_type = self.data.get("test_agent_type","best")

        # Tab Q Agent parameters
        self.alpha = self.data.get("alpha", [0.05])
        self.gamma = self.data.get("gamma", [0.95])
        self.epsilon = self.data.get("epsilon", [0.05])
        # Neural Agent Parameters
        self.input_type = "lm"
        self.input_size = self.data.get("input_size", [384])
        self.sent_hidden_dim = self.data.get("sent_hidden_dim", [10])
        self.hidden_dim = self.data.get("hidden_dim", [128])
        self.num_hidden = self.data.get("num_hidden", [2])
        self.sequence_size = self.data.get("sequence_size", [20])
        self.memory_size = self.data.get("memory_size", [2000])
        self.target_replace_iter = self.data.get("target_replace_iter", [100])
        self.learning_rate = self.data.get("learning_rate", [0.001])
        self.batch_size = self.data.get("batch_size", [1])
        
        self.number_test_episodes = self.data.get("number_test_episodes", 250)
        self.number_test_repeats = self.data.get("number_test_repeats", 5)
        
class TestingSetupConfig(Config):
    def __init__(self, config_dir: str):
        super(TestingSetupConfig, self).__init__(config_dir)
        self.state_configs = ExperimentConfig(os.path.join(config_dir))
                                              
        


class DefaultAgentConfig:
    def __init__(self):
        self.data ={   
            "name": "Default",
            "problem_type": "Default",
            
            "experience_sample_batch_ratio": 0,
             
            "number_training_episodes": 1000,
            "number_training_repeats": 5,
            "number_training_seeds": 1,

            "test_agent_type":"best",
            "number_test_episodes": 200,
            "number_test_repeats": 10,

            "agent_select": ["Qlearntab"],
            "adapter_select": ["default"],
            "agent_parameters":{
                "Qlearntab":{
                    "alpha": 0.1,
                    "gamma": 0.95,
                    "epsilon": 1,
                    "epsilon_step":0
                    },
            }
        }



import time
import numpy as np
import pandas as pd
from typing import List, Dict, Iterable
from abc import ABC, abstractmethod
from elsciRL.adapters import StateAdapter
from torch import Tensor


class Encoder(ABC):
    @abstractmethod
    def encode(self, *args, **kwargs) -> Tensor:
        pass

class StateEncoder(Encoder):
    tensor_cache: Dict[int, Tensor] = dict()
    tensor_cache_index: int = 0

    @staticmethod
    def cache_insert(t: Tensor):
        StateEncoder.tensor_cache[StateEncoder.tensor_cache_index] = t
        StateEncoder.tensor_cache_index += 1
    
    @staticmethod
    def cache_retrieve(offset: int, index: int):
        return StateEncoder.tensor_cache[offset + index]
    
    # index_objects are the complete list of adapter specific elements used to define the encoder's index
    def encode(self, index_objects:list=None, state:list = None, legal_actions:list = None, prior_action:str = None,
                opponent_action:str = None, indexed: bool = False) -> Tensor:
        pass
    
    
class EncodedState(ABC):
    @abstractmethod
    def data() -> Iterable:
        raise NotImplementedError


class StateConverter(ABC):
    def __init__(self, adapter: StateAdapter):
        super().__init__()
        # Calls the conversion procedure
        self.data: EncodedState = self.convert(adapter.s)


    def convert(state: list) -> EncodedState:
        pass


import os
import torch
# ------ Interaction Protocol -----------------------------------
from elsciRL.interaction_loops.standard import StandardInteractionLoop
# ------ Agent Imports -----------------------------------------
from elsciRL.agents.table_q_agent import TableQLearningAgent
from elsciRL.agents.neural_q_agent import NeuralQLearningAgent
from elsciRL.agents.random_agent import RandomAgent
# --------------------------------------------------------------
# TODO use encoder defined by config not manual import
from elsciRL.encoders.sentence_transformer_MiniLM_L6v2 import LanguageEncoder

# TODO: Enable any number of the same agent types with varying parameters
AGENT_TYPES = {
    "Qlearntab": TableQLearningAgent,
    "Neural_Q": NeuralQLearningAgent,
    "Neural_Q_2": NeuralQLearningAgent,
    "Neural_Q_language": NeuralQLearningAgent,
    "Random": RandomAgent
}



class elsciRLSearch:
    def __init__(self, Config:dict, LocalConfig:dict, 
                 Engine, Adapters:dict,
                 save_dir:str, 
                 number_exploration_episodes:int = 10000,
                 match_sim_threshold:float=0.9,
                 observed_states:dict = {}):
        
        # ----- Configs
        self.observed_states = observed_states
        # Meta parameters
        self.ExperimentConfig = Config
        # Local Parameters
        self.ProblemConfig = LocalConfig
        
        self.engine = Engine
        self.adapters = Adapters
        self.env = StandardInteractionLoop
        self.setup_info:dict = self.ExperimentConfig | self.ProblemConfig  
        self.training_setups: dict = {}
        self.instruction_results:dict = {}

        save_dir_extra = save_dir.split("/")[-1]
        save_dir = '/'.join(save_dir.split("/")[:-1])
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)
        self.save_dir = save_dir+'/'+save_dir_extra
        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir) 

        # Unsupervised search parameters
        self.agent_type: str = "Qlearntab" # fix to one type for fast search
        self.epsilon: float = 1 # fully random search agent
        self.number_exploration_episodes: int = number_exploration_episodes

        # Unsupervised search parameters
        self.enc = LanguageEncoder()
        self.sim_threshold: float = match_sim_threshold
        self.cos = torch.nn.CosineSimilarity(dim=0)  
        

    def search(self, action_cap:int=100):
        # Fixed to Tabular Q learning agent for now
        agent_type = self.agent_type
        # We are adding then overriding some inputs for exploration setups
        train_setup_info = self.setup_info.copy()
        # Override action cap for shorter term sub-goals for faster learning
        train_setup_info['training_action_cap'] = action_cap 
        # ----- State Adapter Choice
        for adapter in train_setup_info["adapter_select"]:
            if (("language" in adapter.lower()) |
                ("lang" in adapter.lower()) |
                ("_l" in adapter.lower())|
                ("l_" in adapter.lower())):
                train_setup_info["adapter_select"] = adapter
                break

        #adapter = train_setup_info["adapter_select"][0]
        # ----- Agent parameters
        agent_parameters = train_setup_info["agent_parameters"][agent_type]
        train_setup_info['agent_type'] = agent_type
        train_setup_info['agent_name'] = (str(agent_type) + '_' +
                                        str(adapter) + '_' +
                                        str(agent_parameters))
        train_setup_info['adapter_select'] = adapter
        # ----- init agent
        player = AGENT_TYPES[agent_type](**agent_parameters)
        train_setup_info['agent'] = player
        # -----
        # Set env function to training
        train_setup_info['train'] = True
        # --- 
        # Set exploration parameters
        train_setup_info['number_training_episodes'] = self.number_exploration_episodes 
        train_setup_info['epsilon'] = self.epsilon 
        # ---------------------------------elsciRL-----------------------------------------
        # Train on Live system for limited number of total episodes
        train_setup_info['training_results'] = False
        train_setup_info['observed_states'] = False
        train_setup_info['experience_sampling'] = False
        train_setup_info['live_env'] = True 
        # ---------------------------
        # Init environment to define current position
        train_setup_info['sub_goal'] = None 
        sample_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
        print("Environment Init Position: ", sample_env.start_obs)
        # ---
        # Explore env with limited episodes
        # Environment now init here and called directly in experimental setup loop
        # - setup elsciRL info
        # Train on Live system for limited number of total episodes
        explore_results = sample_env.episode_loop()
        train_setup_info['training_results'] = explore_results
        train_setup_info['observed_states'] = sample_env.elsciRL.observed_states
        train_setup_info['experience_sampling'] = sample_env.elsciRL.experience_sampling
        # Extract visited states from env
        self.observed_states = sample_env.elsciRL.observed_states
        # ---------------------------
        
        return self.observed_states
    
    def match(self, action_cap:int=5, instructions:list=[''], instr_descriptions:list=['']):
        device = "cuda" if torch.cuda.is_available() else "cpu" 
        # Fixed to Tabular Q learning agent for now
        agent_type = self.agent_type
        # We are adding then overriding some inputs for exploration setups
        train_setup_info = self.setup_info.copy()
        # Override action cap for shorter term sub-goals for faster learning
        train_setup_info['training_action_cap'] = action_cap 
        # ----- State Adapter Choice
        adapter = train_setup_info["adapter_select"][0]
        # ----- Agent parameters
        agent_parameters = train_setup_info["agent_parameters"][agent_type]
        train_setup_info['agent_type'] = agent_type
        train_setup_info['agent_name'] = (str(agent_type) + '_' +
                                        str(adapter) + '_' +
                                        str(agent_parameters))
        train_setup_info['adapter_select'] = adapter
        # ----- init agent
        player = AGENT_TYPES[agent_type](**agent_parameters)
        train_setup_info['agent'] = player
        # -----
        # Set env function to training
        train_setup_info['train'] = True
        # ---------------------------------elsciRL-----------------------------------------
        # Train on Live system for limited number of total episodes
        train_setup_info['training_results'] = False
        train_setup_info['observed_states'] = False
        train_setup_info['experience_sampling'] = False
        train_setup_info['live_env'] = True 
        # ---------------------------
        # Init environment to define current position
        train_setup_info['sub_goal'] = None 
        sample_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
            
        # New: user input here
        #instructions, instr_descriptions = self.elsciRL_input.user_input()
        # DEMO SETS THIS AS FUNCTION INPUT
        # ---------------------------
        best_match_results = {}
        for i,instr in enumerate(instructions):
            if i == 0:  
                # env start is '0' so force rest to start from '1' instead otherwise 
                # breaks uniqueness requirement for instructions
                instruction = str(sample_env.start_obs).split(".")[0] + "---" + str(int(instr)+1)
            else:
                instruction = str(int(instructions[i-1])+1) + "---" + str(int(instr)+1)
            instr_description = instr_descriptions[i]
            
            if type(instr_description) == type(''):
                instr_description = instr_description.split('.')
                instr_description = list(filter(None, instr_description))
            # Create tensor vector of description
            instruction_vector = self.enc.encode(instr_description)
            # Default fedeback layer - DEMO wont currently updated this 
            feedback_layer = torch.zeros(instruction_vector.size()).to(device)
            # EXPLORE TO FIND LOCATION OF SUB-GOAL
            sub_goal = None
            # ---------------------------
            if (instruction in self.instruction_results):
                if (agent_type+'_'+adapter) in self.instruction_results[instruction]:
                    # We use feedback layer even if sub_goal not a good match
                    feedback_layer = self.instruction_results[instruction][agent_type+'_'+adapter]['feedback_layer']
                    if (self.instruction_results[instruction][agent_type+'_'+adapter]['sim_score']>=self.sim_threshold):
                        sub_goal = self.instruction_results[instruction][agent_type+'_'+adapter]['sub_goal'][0]
                        sub_goal_list = self.instruction_results[instruction][agent_type+'_'+adapter]['sub_goal']
                        sim = self.instruction_results[instruction][agent_type+'_'+adapter]['sim_score']
                else:
                    self.instruction_results[instruction][agent_type+'_'+adapter] = {}
                    feedback_layer = torch.zeros(instruction_vector.size()).to(device)
                self.instruction_results[instruction][agent_type+'_'+adapter]['count'] = self.instruction_results[instruction][agent_type+'_'+adapter]['count']+1
            else:
                self.instruction_results[instruction] = {}    
                self.instruction_results[instruction][agent_type+'_'+adapter] = {} 
                self.instruction_results[instruction][agent_type+'_'+adapter]['count'] = 1
                self.instruction_results[instruction][agent_type+'_'+adapter]['action_cap'] = action_cap
                feedback_layer = torch.zeros(instruction_vector.size()).to(device)
                self.instruction_results[instruction][agent_type+'_'+adapter]['feedback_layer'] = feedback_layer
            # ---------------------------
            while not sub_goal:
                # If no description -> no sub-goal (i.e. envs terminal goal position)
                if not instr_description: 
                    sub_goal = None
                    # If no sub-goal -> find best match of description from env 
                else:
                    # Compare to instruction vector                            
                    max_sim = -1
                    # all states that are above threshold 
                    sub_goal_list = []
                    for obs_state in self.observed_states:
                        str_state = self.observed_states[obs_state]
                        t_state = self.enc.encode(str_state)
                        # ---
                        total_sim = 0
                        dim_count = 0 # For some reason encoder here is adding extra dimension                                        
                        for idx,instr_sentence in enumerate(instruction_vector):
                            feedback_layer_sent = feedback_layer[idx]
                            for state_sentence in t_state:
                                total_sim+=self.cos(torch.add(state_sentence, feedback_layer_sent), instr_sentence)
                                dim_count+=1
                       
                        sim = 0 if dim_count==0 else total_sim.item()/dim_count
                        if sim > max_sim:
                            max_sim  = sim
                            sub_goal_max = obs_state
                            #sub_goal_max_t = t_state
                        if sim >= self.sim_threshold:
                            sub_goal = obs_state # Sub-Goal code
                            sub_goal_max = obs_state
                            #sub_goal_t = t_state
                            sub_goal_list.append(sub_goal)
    
                    # OR if none above threshold matching max sim
                    if max_sim < self.sim_threshold:
                        sub_goal = sub_goal_max
                        #sub_goal_t = sub_goal_max_t                                    
                        # Find all states that have same sim as max
                        for obs_state in self.observed_states:
                            str_state = self.observed_states[obs_state]
                            t_state = self.enc.encode(str_state)
                            # ---
                            total_sim = 0
                            # Average sim across each sentence in instruction vs state
                            dim_count = 0
                            for idx,instr_sentence in enumerate(instruction_vector):
                                feedback_layer_sent = feedback_layer[idx]
                                for state_sentence in t_state:
                                    total_sim+=self.cos(torch.add(state_sentence, feedback_layer_sent), instr_sentence)
                                    dim_count+=1
                            sim = 0 if dim_count==0 else total_sim.item()/dim_count
                            if sim >= (max_sim):
                                sub_goal_list.append(obs_state)

                    if max_sim < self.sim_threshold:
                        print("Minimum sim for observed states to match instruction not found, using best match instead. Best match sim value = ", max_sim )
                # If adapter is poor to match to instruction vector none of them observed states match
                if (max_sim<-1)|(sub_goal is None):#|(max_sim>1):
                    print("All observed states result in similarity outside bounds (i.e. not found).")
                    print("Max sim found = ", max_sim)
                    sub_goal_list = ['']
                    break

            if (agent_type+'_'+adapter) in self.instruction_results[instruction]:
                self.instruction_results[instruction][agent_type+'_'+adapter]['sub_goal'] = sub_goal_list
                self.instruction_results[instruction][agent_type+'_'+adapter]['sim_score'] = sim
                self.instruction_results[instruction][agent_type+'_'+adapter]['feedback_layer'] = feedback_layer
            else:
                self.instruction_results[instruction][agent_type+'_'+adapter] = {}
                self.instruction_results[instruction][agent_type+'_'+adapter]['count'] = 1
                self.instruction_results[instruction][agent_type+'_'+adapter]['action_cap'] = action_cap
                self.instruction_results[instruction][agent_type+'_'+adapter]['sub_goal'] = sub_goal_list
                self.instruction_results[instruction][agent_type+'_'+adapter]['sim_score'] = sim
                self.instruction_results[instruction][agent_type+'_'+adapter]['feedback_layer'] = feedback_layer

            best_match_results[instruction] = {}
            best_match_results[instruction]['sub_goal'] = self.observed_states[sub_goal]
            best_match_results[instruction]['best_match'] = sub_goal_max
                       
        return best_match_results, self.instruction_results



class elsciRL_hierarchy_functions:
    """TODO: Specify the functions used to complete the instruction following methods."""
    def __init__(self):
        device = "cuda" if torch.cuda.is_available() else "cpu" 
        # Trigger re-search
        if re_search_override:
            self.observed_states:dict = {}
        pass


    def agent_loop(self):
        for n, agent_type in enumerate(self.setup_info['agent_select']): 
            # We are adding then overriding some inputs from general configs for experimental setups
            train_setup_info = self.setup_info.copy()
            # Override action cap for shorter term sub-goals for faster learning
            train_setup_info['training_action_cap'] = action_cap 
            # ----- State Adapter Choice
            adapter = train_setup_info["adapter_select"][n]
            # ----- Agent parameters
            agent_parameters = train_setup_info["agent_parameters"][agent_type]
            train_setup_info['agent_type'] = agent_type
            train_setup_info['agent_name'] = str(agent_type) + '_' + str(adapter) + '_' + str(agent_parameters)
            train_setup_info['adapter_select'] = adapter
            # ----- init agent
            player = AGENT_TYPES[agent_type](**agent_parameters)
            train_setup_info['agent'] = player
            # -----
            # Set env function to training# Repeat training
            train_setup_info['train'] = True
            # --- 
            # Set exploration parameters
            train_setup_info['number_training_episodes'] = self.number_exploration_episodes # Override
            
            # ---------------------------------elsciRL-----------------------------------------
            # Train on Live system for limited number of total episodes
            train_setup_info['training_results'] = False
            if not self.observed_states:
                train_setup_info['observed_states'] = False
            else:
                train_setup_info['observed_states'] = self.observed_states
            train_setup_info['experience_sampling'] = False
            train_setup_info['live_env'] = True 
            # ---------------------------
            # Init environment to define current position
            self.plan_loop()

    def plan_loop(self):
        for p in range(0, self.num_plans):
            print("---")
            print("Plan Number = ", p)
            sample_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
            print("Environment Init Position: ", sample_env.start_obs)
            # New: user input here    
            instructions, instr_descriptions = self.elsciRL_input.user_input()
            
            feedback_count = 0
            self.repeat_loop()
            # log results of feedback loop
            if (agent_type+'_'+adapter) not in self.feedback_results:
                self.feedback_results[agent_type+'_'+adapter] = {}
        
            if p not in self.feedback_results[(agent_type+'_'+adapter)]:
                self.feedback_results[(agent_type+'_'+adapter)][p] = {}
                self.feedback_results[(agent_type+'_'+adapter)][p]['instr_count'] = i+1
                self.feedback_results[(agent_type+'_'+adapter)][p]['feedback_count'] = feedback_count
            else:
                self.feedback_results[(agent_type+'_'+adapter)][p]['feedback_count'] = feedback_count + self.feedback_results[(agent_type+'_'+adapter)][p]['feedback_count']

    def repeat_loop(self):
        for repeat in range(0,self.feedback_repeats): # Arbitrary repeat to further reinforce matching state to instr
            print("===")
            print("REINFORCEMENT Repeated search num ", repeat+1)
            
            self.instr_loop()

    def instr_loop(self):
        for i,instr in enumerate(instructions):        
            if i == 0:
                instruction = str(sample_env.start_obs).split(".")[0] + "---" + instr
            else:
                instruction = instructions[i-1] + "---" + instr
            
            instr_description = instr_descriptions[i]
            
            if type(instr_description) == type(''):
                instr_description = instr_description.split('.')
                instr_description = list(filter(None, instr_description))
            # Create tensor vector of description
            instruction_vector = self.enc.encode(instr_description)
            print("Size of Encoded Instruction Tensor")
            print(instruction_vector.size())
            # EXPLORE TO FIND LOCATION OF SUB-GOAL
            sub_goal = None
            # ---------------------------
            # Overrides on prior knowledge
            # Seen sub_goal before & sim above threshold
            if (instruction in self.instruction_results):
                if (agent_type+'_'+adapter) in self.instruction_results[instruction]:
                    # We use feedback layer even if sub_goal not a good match
                    feedback_layer = self.instruction_results[instruction][agent_type+'_'+adapter]['feedback_layer']
                    if (self.instruction_results[instruction][agent_type+'_'+adapter]['sim_score']>=self.sim_threshold):
                        sub_goal = self.instruction_results[instruction][agent_type+'_'+adapter]['sub_goal'][0]
                        sub_goal_list = self.instruction_results[instruction][agent_type+'_'+adapter]['sub_goal']
                        sim = self.instruction_results[instruction][agent_type+'_'+adapter]['sim_score']
                else:
                    self.instruction_results[instruction][agent_type+'_'+adapter] = {}
                    feedback_layer = torch.zeros(instruction_vector.size()).to(device)
                self.instruction_results[instruction][agent_type+'_'+adapter]['count'] = self.instruction_results[instruction][agent_type+'_'+adapter]['count']+1
            else:
                self.instruction_results[instruction] = {}    
                self.instruction_results[instruction][agent_type+'_'+adapter] = {} 
                self.instruction_results[instruction][agent_type+'_'+adapter]['count'] = 1
                self.instruction_results[instruction][agent_type+'_'+adapter]['action_cap'] = action_cap
                feedback_layer = torch.zeros(instruction_vector.size()).to(device)
                self.instruction_results[instruction][agent_type+'_'+adapter]['feedback_layer'] = feedback_layer
            # ---------------------------
            search_count=0
            while not sub_goal:
                sim_delta = 1 # Parameter that stops updates when change to feedback is small
                # If no description -> no sub-goal (i.e. envs terminal goal position)
                if not instr_description: 
                    sub_goal = None
                    # If no sub-goal -> find best match of description from env 
                else:
                    search_count+=1
                    print("------") 
                    print("Search: ", search_count)
                    
                    # Only run on live env if observed states empty
                    if not self.observed_states:
                        train_setup_info['sub_goal'] = sub_goal # None
                        # ---
                        # Explore env with limited episodes
                        # Environment now init here and called directly in experimental setup loop
                        # - setup elsciRL info
                        # Train on Live system for limited number of total episodes
                        train_setup_info['live_env'] = True                        
                        live_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
                        explore_results = live_env.episode_loop()
                        train_setup_info['training_results'] = explore_results
                        train_setup_info['observed_states'] = live_env.elsciRL.observed_states
                        train_setup_info['experience_sampling'] = live_env.elsciRL.experience_sampling
                        # Extract visited states from env
                        self.observed_states = live_env.elsciRL.observed_states

                    # Compare to instruction vector                            
                    max_sim = -1
                    # all states that are above threshold 
                    sub_goal_list = []
                    for obs_state in self.observed_states:
                        str_state = self.observed_states[obs_state]
                        #str_state_stacked = ' '.join(str_state)
                        t_state = self.enc.encode(str_state)
                        # ---
                        total_sim = 0
                        dim_count = 0 # For some reason encoder here is adding extra dimension                                        
                        for idx,instr_sentence in enumerate(instruction_vector):
                            feedback_layer_sent = feedback_layer[idx]
                            for state_sentence in t_state:
                                total_sim+=self.cos(torch.add(state_sentence, feedback_layer_sent), instr_sentence)
                                dim_count+=1
                        #print("Observation: ", str_state)
                        #print("Tensor check: ", t_state.size(), " --- ", instruction_vector.size(), " --- ", feedback_layer.size())
                        #print("Sim result: ", total_sim.item(), '---', dim_count)
                        
                        sim = 0 if dim_count==0 else total_sim.item()/dim_count
                        if sim > max_sim:
                            max_sim  = sim
                            sub_goal_max = obs_state
                            sub_goal_max_t = t_state
                        if sim >= self.sim_threshold:
                            sub_goal = obs_state # Sub-Goal code
                            sub_goal_t = t_state
                            sub_goal_list.append(sub_goal)
    
                    # OR if none above threshold matching max sim
                    if max_sim < self.sim_threshold:
                        sub_goal = sub_goal_max
                        sub_goal_t = sub_goal_max_t                                    
                        # Find all states that have same sim as max
                        for obs_state in self.observed_states:
                            str_state = self.observed_states[obs_state]
                            #str_state_stacked = ' '.join(str_state)
                            t_state = self.enc.encode(str_state)
                            # ---
                            total_sim = 0
                            # Average sim across each sentence in instruction vs state
                            dim_count = 0
                            for idx,instr_sentence in enumerate(instruction_vector):
                                feedback_layer_sent = feedback_layer[idx]
                                for state_sentence in t_state:
                                    total_sim+=self.cos(torch.add(state_sentence, feedback_layer_sent), instr_sentence)
                                    dim_count+=1
                            sim = 0 if dim_count==0 else total_sim.item()/dim_count
                            if sim >= (max_sim):
                                sub_goal_list.append(obs_state)

                    if max_sim < self.sim_threshold:
                        print("Minimum sim for observed states to match instruction not found, using best match instead. Best match sim value = ", max_sim )
                # If adapter is poor to match to instruction vector none of them observed states match
                if (max_sim<-1)|(sub_goal is None):#|(max_sim>1):
                    print("All observed states result in similarity outside bounds (i.e. strongly opposite vectors to instruction). Re-starting Search.")
                    print("Max sim found = ", max_sim)
                    sub_goal = None
                    self.observed_states = {}
                elif (sim_delta<0.05)|(search_count>=3):
                    print("Change in sim less than or equal to delta cap or search cap reached, assume goal-state not observed. Re-starting Search.")
                    if simulated_instr_goal:
                        print("-- Known sub_goal position: ", simulated_instr_goal)
                    print("-- Best match: ", sub_goal_max)
                    sub_goal = None
                    self.observed_states = {}
                    search_count = 0
                else:
                    if not simulated_instr_goal:
                        
                        for s_g in sub_goal_list:
                            print(self.observed_states[s_g], ": ", s_g)
                        print(" ")
                        print("===========")
                        print("Instruction: ", instr)
                        print(sub_goal)
                        print("Best match state for instruction \n ______ \n Adapted form: \n\t - ", self.observed_states[sub_goal], " \n Engine observation: \n\t - " , sub_goal)
                        print("Full list of matching states...",)
                        print("")
                        print("Tensor check: Size of Instruction vs Best match: ", instruction_vector.size(), " --- ", sub_goal_t.size())
                        print("")
                        feedback = input("-- Does this match the expectation instruction outcome? (Y/N)")
                        feedback_count+=1 
                    else:
                        if type(simulated_instr_goal[0]) != type(sub_goal_max):
                            print("- ERROR: Typing of simulated sub-goal check does not match typing of state from environment, please correct this.")
                            print(type(simulated_instr_goal[0])," - ", type(sub_goal_max))
                            #print("Observed States Examples: ", self.observed_states(list(self.observed_states.keys())[0:5]))
                            exit()
                        else:
                            if sub_goal_max in simulated_instr_goal:
                                print("- Simulated sub-goal match found.")
                                print("-- Known sub_goal position: ", simulated_instr_goal)
                                print("-- Unsupervised state best match: ", sub_goal_max)
                                feedback = 'Y'
                            else:
                                print("- Match NOT found.")
                                print("-- Known sub_goal position: ", simulated_instr_goal)
                                print("-- Unsupervised state best match: ", sub_goal_max)
                                feedback = 'N'
                    
                    if (feedback.lower() == 'y')|(feedback.lower() == 'yes'):
                        for idx,instr_sentence in enumerate(instruction_vector):
                            feedback_layer_sent = feedback_layer[idx]
                            for sentence in sub_goal_t:
                                feedback_layer[idx] = torch.add(feedback_layer_sent, self.feedback_increment*(torch.sub(instr_sentence, sentence))) 
                        total_sim = 0
                        # Average sim across each sentence in instruction vs state
                        dim_count = 0
                        for idx,instr_sentence in enumerate(instruction_vector):
                            feedback_layer_sent = feedback_layer[idx]
                            for state_sentence in t_state:
                                total_sim+=self.cos(torch.add(state_sentence, feedback_layer_sent), instr_sentence)
                                dim_count+=1
                        sim = 0 if dim_count==0 else total_sim.item()/dim_count
                        sim_delta = sim-max_sim
                        print("--- Change in sim results with POSITIVE reinforcement of correct state match =", sim_delta)
                        print("--- New Sim Value = ",  sim)
                    else:
                        for idx,instr_sentence in enumerate(instruction_vector):
                            feedback_layer_sent = feedback_layer[idx]
                            for sentence in sub_goal_t:
                                feedback_layer[idx] = torch.sub(feedback_layer_sent, self.feedback_increment*(torch.sub(instr_sentence, sentence)))
                        total_sim = 0
                        # Average sim across each sentence in instruction vs state
                        dim_count = 0
                        for idx,instr_sentence in enumerate(instruction_vector):
                            feedback_layer_sent = feedback_layer[idx]
                            for state_sentence in t_state:
                                total_sim+=self.cos(torch.add(state_sentence, feedback_layer_sent), instr_sentence)
                                dim_count+=1
                        sim = 0 if dim_count==0 else total_sim.item()/dim_count
                        sim_delta = sim-max_sim
                        print("--- Change in sim results with NEGATIVE reinforcement for NO MATCH =", sim_delta)
                        sub_goal = None
            if (agent_type+'_'+adapter) not in self.instruction_results[instruction]:
                self.instruction_results[instruction][agent_type+'_'+adapter] = {}   
            self.instruction_results[instruction][agent_type+'_'+adapter]['sub_goal'] = sub_goal_list
            self.instruction_results[instruction][agent_type+'_'+adapter]['sim_score'] = sim
            self.instruction_results[instruction][agent_type+'_'+adapter]['feedback_layer'] = feedback_layer


        


import os
import json

# ------ Experiment Import --------------------------------------
from elsciRL.evaluation.standard_report import Evaluation

# ------ Agent Imports -----------------------------------------
# Universal Agents
from elsciRL.agents.agent_abstract import Agent, QLearningAgent
from elsciRL.agents.table_q_agent import TableQLearningAgent
from elsciRL.agents.neural_q_agent import NeuralQLearningAgent

AGENT_TYPES = {
    "Qlearntab": TableQLearningAgent,
    "Neural_Q": NeuralQLearningAgent,
    "Neural_Q_2": NeuralQLearningAgent,
    "Neural_Q_language": NeuralQLearningAgent
}

PLAYER_PARAMS = {
    "Qlearntab": ["alpha", "gamma", "epsilon"],
    "Neural_Q": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "Neural_Q_2": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "Neural_Q_language": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"]
}

# This is the main run functions for elsciRL to be imported
# Defines the train/test operators and imports all the required agents and experiment functions ready to be used
# The local main.py file defines the [adapters, configs, environment] to be input

# TODO: This should be where the environment is initialized and then episode_loop (or train/test) is run
# -> results then passed down to experiment to produce visual reporting (staticmethod)
# -> instruction following approach then becomes alternative form of this file to be called instead
# -> TODO: This means we have multiple main.py types (e.g. with/without convergence measure) so should create a directory and finalise naming for this

class SupervisedExperiment:
    def __init__(self, Config:dict, LocalConfig:dict, Environment, 
                 save_dir:str, show_figures:str, window_size:float, 
                 instruction_path: dict, predicted_path: dict=None):
        self.ExperimentConfig = Config
        self.LocalConfig = LocalConfig
        
        if not predicted_path:
            if not os.path.exists(save_dir):
                os.mkdir(save_dir)
            self.save_dir = save_dir+'/Supervised_Instr_Experiment'
        else:
            save_dir_extra = save_dir.split("/")[-1]
            save_dir = '/'.join(save_dir.split("/")[:-1])
            if not os.path.exists(save_dir):
                os.mkdir(save_dir)
            self.save_dir = save_dir+'/'+save_dir_extra

        self.show_figures = show_figures
        if (self.show_figures.lower() != 'y')|(self.show_figures.lower() != 'yes'):
            print("Figures will not be shown and only saved.")

        self.env = Environment
        self.setup_info = self.ExperimentConfig['data'] | vars(self.LocalConfig) # TODO: configs aren't consistent formatting 
        self.training_setups: dict = {}
        # New instruction learning
        # New for when we have predicted goal states for training
        if predicted_path:
            self.instruction_path = predicted_path
            if instruction_path:
                self.testing_path = instruction_path
            else:
                self.testing_path = predicted_path
        else:
            self.instruction_path = instruction_path
            self.testing_path = False
        self.instruction_agents:dict = {}
    
        self.analysis = Evaluation(window_size=window_size)
        
    def train(self):
        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir)

        prior_sub_goal = None
        for i,instruction in enumerate(list(self.instruction_path.keys())): 
            print(" ")
            print("Instruction Sub-Goal: ", instruction)
            instr_save_dir = self.save_dir+'/'+str(i+1)+"-"+instruction.replace(" ","")
            self.instr_dir = instr_save_dir
            if not os.path.exists(instr_save_dir):
                os.mkdir(instr_save_dir)
                
            for n, agent_type in enumerate(self.setup_info['agent_select']):
                # We are adding then overriding some inputs from general configs for experimental setups
                train_setup_info = self.setup_info.copy()
                train_setup_info['instruction'] = instruction
                # Override action cap for shorter term sub-goals for faster learning
                if 'action_cap' in self.instruction_path[instruction]:
                    train_setup_info['training_action_cap'] = self.instruction_path[instruction]['action_cap'] 
                # ----- State Adapter Choice
                adapter = train_setup_info["adapter_select"][n]
                # ----- Agent parameters
                agent_parameters = train_setup_info["agent_parameters"][agent_type]
                train_setup_info['agent_type'] = agent_type
                train_setup_info['agent_name'] = str(agent_type) + '_' + str(adapter) + '_' + str(agent_parameters)
                train_setup_info['adapter_select'] = adapter
                # ----- Sub-Goal
                # - If we have setup dict to include agent_adapter specific location of sub-goals
                #   i.e. {instr:{env_code:{agent_adapter:{sub_goal:'ENV_CODE', sim_score:0.8}}, action_cap:5}}
                #   Otherwise is standard user defined input {instr:{env_code:'ENV_CODE', action_cap:5}}
                
                # If unsupervised/reinf search not performed then dict structure wont include agent_type/adapter
                if type(self.instruction_path[instruction]['env_code']) == type({}):
                    sub_goal = self.instruction_path[instruction]['env_code'][agent_type+'_'+adapter]['sub_goal']
                    train_setup_info['sub_goal'] = sub_goal
                else:
                    sub_goal = self.instruction_path[instruction]['env_code']
                    train_setup_info['sub_goal'] = sub_goal
                # -----
                # Repeat training
                train_setup_info['train'] = True
                number_training_episodes = train_setup_info['number_training_episodes']
                number_training_repeats = self.ExperimentConfig["number_training_repeats"]
                print("Training Agent " + str(agent_type) + " for " + str(number_training_repeats) + " repeats")
                
                temp_agent_store = {}
                setup_num:int = 0
                for training_repeat in range(1,(number_training_repeats+1)):
                    print("------")
                    print("- Repeat Num: ", training_repeat)
                    setup_num+=1
                    # ----- init agent
                    if instruction not in self.instruction_agents:
                        # Instructions are in sequence so we initialise later instructions with policy of one prior
                        if not prior_sub_goal:
                            player = AGENT_TYPES[agent_type](**agent_parameters)
                        else:
                            self.instruction_agents[prior_instruction][agent_type].clone() # We need it to adopt prior sub-goal but create a new instance otherwise repeats wont reset learning
                            #player = AGENT_TYPES[agent_type](**agent_parameters)
                            #player.load(self.instruction_agents[prior_sub_goal][agent_type])  
                    else:
                        self.instruction_agents[instruction][agent_type].clone()
                        #player = AGENT_TYPES[agent_type](**agent_parameters)
                        #player.load(self.instruction_agents[sub_goal][agent_type])
                    train_setup_info['agent'] = player
                    # -------------------------------------------------------------------------------
                    # Initialise Environment
                    # Environment now init here and called directly in experimental setup loop

                    # Initialise Environment
                    # Environment now init here and called directly in experimental setup loop
                    train_setup_info['training_results'] = False
                    train_setup_info['observed_states'] = False
                    train_setup_info['experience_sampling'] = False
                    if train_setup_info['experience_sample_batch_ratio']>0:
                        live_sample_batch_size = int(number_training_episodes*train_setup_info['experience_sample_batch_ratio'])
                        live_sample_batch_count = int(1/train_setup_info['experience_sample_batch_ratio'])
                        print("-- Training with Simulated Batches, ", live_sample_batch_count, " total...")
                        for live_sample_batch in range(0, live_sample_batch_count-1):
                            print("--- Live Interaction Batch Num", live_sample_batch+1)
                            # Train on Live system for limited number of total episodes
                            train_setup_info['live_env'] = True                        
                            train_setup_info['number_training_episodes'] = live_sample_batch_size
                            live_env = self.env(train_setup_info)
                            training_results = live_env.episode_loop()

                            # Update shared information between init environments
                            train_setup_info['training_results'] = training_results
                            train_setup_info['observed_states'] = live_env.elsciRL.observed_states
                            train_setup_info['experience_sampling'] = live_env.elsciRL.experience_sampling
                            # Train based on simulated exp
                            train_setup_info['live_env'] = False
                            train_setup_info['number_training_episodes'] = number_training_episodes

                            simulated_env = self.env(train_setup_info)
                            simulated_env.episode_loop()
                        # Final batch doesn't require simulated exp after and we need result output
                        print("--- Final batch")
                        train_setup_info['live_env'] = True
                        train_setup_info['number_training_episodes'] = live_sample_batch_size
                        training_results = live_env.episode_loop()
                        # Have to 'fix' output
                        training_results['episode'] = training_results.index
                        cumulative_r = 0
                        cumulative_r_lst = []
                        for r in training_results['episode_reward']:
                            cumulative_r+=r
                            cumulative_r_lst.append(cumulative_r)
                        training_results['cumulative_reward'] = cumulative_r_lst
                    else:
                        train_setup_info['live_env'] = True
                        live_env = self.env(train_setup_info)
                        training_results = live_env.episode_loop()

                    # Opponent now defined in local setup.py
                    # ----- Log training setup      
                    training_results.insert(loc=0, column='Repeat', value=setup_num)
                    # Extract trained agent from env
                    agent = live_env.agent
                    # ----- Save Directory -> Sub-directory for each setup type (parameters+opponents+repeats)
                    # -> I also moved env init into inner most loop because I'm worried the agent isn't getting reset otherwise
                    agent_save_dir = instr_save_dir+'/'+agent_type+'_'+adapter+'_'+str(setup_num) 
                    
                    if not os.path.exists(agent_save_dir):
                        os.mkdir(agent_save_dir)
                    # Produce training report with Analysis.py
                    Return = self.analysis.train_report(training_results, agent_save_dir, self.show_figures)
                                        
                    # Save trained agent to logged output
                    train_setup_info['train_save_dir'] = agent_save_dir
                    train_setup_info['trained_agent'] = agent
                    # Collate complete setup info to full dict
                    self.training_setups['Training_Setup_'+str(agent_type) + '_' + str(adapter)+'_'+str(setup_num)] = train_setup_info.copy()
                    temp_agent_store[setup_num] = {'Return':Return,'agent':agent}

                # Only save the best agent from repeated training
                best_return = float('-inf')
                best_agent = None
                for repeat in range(1, training_repeat + 1):
                    if temp_agent_store[repeat]['Return'] > best_return:
                        best_return = temp_agent_store[repeat]['Return']
                        best_agent = temp_agent_store[repeat]['agent']
                        print(best_return)

                if instruction not in self.instruction_agents:
                    self.instruction_agents[instruction] = {}
                self.instruction_agents[instruction][agent_type] = best_agent#.save()          
            # For next instruction to adopt policy from
            prior_instruction = instruction
            prior_sub_goal = sub_goal
            if number_training_repeats>1:
                self.analysis.training_variance_report(instr_save_dir, self.show_figures)

        #json.dump(self.training_setups) # TODO: Won't currently serialize this output to a json file
        print(self.instruction_agents) 
        return self.training_setups

    # TESTING PLAY
    def test(self, training_setups:str=None):
        # Override input training setups with previously saved 
        if training_setups is None:
            training_setups = self.training_setups
        else:
            json.load(training_setups)

        for training_key in list(training_setups.keys()):    
            test_setup_info = training_setups[training_key]
            test_setup_info['train'] = False # Testing Phase
            print("----------")
            print("Testing results for trained agents in saved setup configuration:")
            print(test_setup_info['train_save_dir'])
            # Extract trained agent
            test_setup_info['agent'] = test_setup_info['trained_agent'] #Update init agent to pre-trained
            number_training_repeats = test_setup_info['number_test_repeats']

            # If we trained with predicted goal positions we now test with correct, supervised goals
            if self.testing_path:
                instruction = test_setup_info['instruction']
                test_setup_info['sub_goal'] = self.testing_path[instruction]['env_code']

            for testing_repeat in range(0, test_setup_info['number_test_repeats']):  
                # Re-init env for testing
                env = self.env(test_setup_info)
                # Testing generally is the agents replaying on the testing ENV
                testing_results = env.episode_loop() 
                test_save_dir = test_setup_info['train_save_dir']+'/testing_results_'+str(testing_repeat)
                if not os.path.exists(test_save_dir):
                    os.mkdir(test_save_dir)
                # Produce training report with Analysis.py
                Return = self.analysis.test_report(testing_results, test_save_dir, self.show_figures)

        # Path is the experiment save dir + the final instruction
        if number_training_repeats>1:
            self.analysis.testing_variance_report(self.instr_dir, self.show_figures)


from openai import OpenAI

def call_gpt_api(prompt):
    import os
    api_key = os.environ.get("OPENAI_API_KEY")
    client = OpenAI(api_key=api_key)
    response = client.chat.completions.create(
        model="gpt-4.1",
        messages=[{"role": "system", "content": prompt}],
        max_tokens=5000
    )
    return response.to_dict() if hasattr(response, 'to_dict') else response

def process_gpt_response(response):
    if response and 'choices' in response:
        return response['choices'][0]['message']['content']
    return None


# TODO: Simplify and remove sub-goals/elsciRL tracking/live_env/exp sampling
import time
import numpy as np
from tqdm import tqdm
import io
from PIL import Image  # Used to generate GIF

# ------ Imports -----------------------------------------
# Agent Setup
from elsciRL.environment_setup.imports import ImportHelper

# Evaluation standards
from elsciRL.environment_setup.results_table import ResultsTable
from elsciRL.environment_setup.elsciRL_info import elsciRLInfo


class StandardInteractionLoop:
    """Interaction Loop for standard environments.
    REQUIRES:
        - Engine: Environment engine defined with elsciRLAI format
        - Adapters: Dictionary of local adapters with unique names: {"name_1": Adapter_1, "name_2": Adapter_2,...}
        - local_setup_info: Dictionary of local setup info (i.e. local config file)
    """
    @staticmethod
    def fig2img(fig):
        """Convert a Matplotlib figure to a PIL Image and return it"""
        buf = io.BytesIO()
        fig.savefig(buf)
        buf.seek(0)
        img = Image.open(buf)
        return img



    def __init__(self, Engine, Adapters: dict, local_setup_info: dict):
        # --- INIT state space from engine
        self.agent_adapter_name = local_setup_info['agent_type'] + "_" + local_setup_info['adapter_select']
        self.engine = Engine(local_setup_info)
        self.start_obs = self.engine.reset()
        # if self.engine.render() is not None:
        #     self.render = self.engine.render()
        # else:
        #     self.render = None
        # ---
        # --- PRESET elsciRL INFO
        # Agent
        Imports = ImportHelper(local_setup_info)
        self.agent, self.agent_type, self.agent_name, self.agent_state_adapter = (
            Imports.agent_info(Adapters)
        )
        (
            self.num_train_episodes,
            self.num_test_episodes,
            self.training_action_cap,
            self.testing_action_cap,
            self.reward_signal,
        ) = Imports.parameter_info()
        # Training or testing phase flag
        self.train = Imports.training_flag()
        # --- elsciRL
        self.live_env, self.observed_states, self.experience_sampling = (
            Imports.live_env_flag()
        )
        # Results formatting
        self.results = ResultsTable(local_setup_info)
        # elsciRL input function
        # - We only want to init trackers on first batch otherwise it resets knowledge
        self.elsciRL = elsciRLInfo(self.observed_states, self.experience_sampling)
        # Env start position for instr input
        # Enable sub-goals
        if (
            (local_setup_info["sub_goal"] is not None)
            & (local_setup_info["sub_goal"] != ["None"])
            & (local_setup_info["sub_goal"] != "None")
        ):
            self.sub_goal: list = local_setup_info["sub_goal"]
        else:
            self.sub_goal: list = None
        # New: multi sub-goal paths used for continuous instruction plans
        # - Would not be defined by config files but rather experiment module so dont need to define in terms of local_setup_info
        self.multi_sub_goal: dict = None

    def episode_loop(self, render: bool = False, render_save_dir: str = None):
        # Mode selection (already initialized)
        if self.train:
            number_episodes = self.num_train_episodes
        else:
            number_episodes = self.num_test_episodes

        sub_goal_tracker = (
            {}
        )  # Used to track sub-goal completion -> re-call most common end state for start of next with EXACT chaining
        sub_goal_best = 0
        self.sub_goal_end = None
        # NEW: Animation of Learning
        if render:
            render_stack = []
        print("\n Episode Interaction Loop: ")
        for episode in tqdm(range(0, number_episodes)):
            action_history = []
            # ---
            # Start observation is used instead of .reset()  fn so that this can be overridden for repeat analysis from the same start pos
            obs = self.engine.reset(start_obs=self.start_obs)
            if render:
                if isinstance(self.engine.render(), np.ndarray):
                    render_stack.append(Image.fromarray(self.engine.render().astype("uint8")))
                else:
                    img = self.fig2img(self.engine.render())
                    render_stack.append(img)
            legal_moves = self.engine.legal_move_generator(obs)
            state = self.agent_state_adapter.adapter(
                state=obs,
                legal_moves=legal_moves,
                episode_action_history=action_history,
                encode=True,
            )
            # ---
            start_time = time.time()
            episode_reward: int = 0
            # ---
            sub_goal_reward_obtained = False
            multi_sub_goal_tracker = (
                []
            )  # Used to tracked known sub-goal completion and reduce print clutter
            multi_goal_reward_obtained = []

            for action in range(0, self.training_action_cap):
                if self.live_env:
                    # Agent takes action
                    legal_moves = self.engine.legal_move_generator(obs)
                    agent_action = self.agent.policy(state, legal_moves)

                    if isinstance(agent_action, np.int64):
                        action_history.append(agent_action.item())
                    else:
                        action_history.append(agent_action)

                    next_obs, reward, terminated, _ = self.engine.step(
                        state=obs, action=agent_action
                    )
                    if render:
                        if isinstance(self.engine.render(), np.ndarray):
                            render_stack.append(Image.fromarray(self.engine.render().astype("uint8")))
                        else:
                            img = self.fig2img(self.engine.render())
                            render_stack.append(img)
                    # Can override reward per action with small negative punishment
                    if reward == 0:
                        reward = self.reward_signal[1]

                    legal_moves = self.engine.legal_move_generator(next_obs)
                    next_state = self.agent_state_adapter.adapter(
                        state=next_obs,
                        legal_moves=legal_moves,
                        episode_action_history=action_history,
                        encode=True,
                    )
                    # elsciRL trackers
                    self.elsciRL.observed_state_tracker(
                        engine_observation=next_obs,
                        language_state=self.agent_state_adapter.adapter(
                            state=next_obs,
                            legal_moves=legal_moves,
                            episode_action_history=action_history,
                            encode=False,
                        ),
                    )

                    # MUST COME BEFORE SUB-GOAL CHECK OR 'TERMINAL STATES' WILL BE FALSE
                    self.elsciRL.experience_sampling_add(
                        state, agent_action, next_state, reward, terminated
                    )
                    # Trigger end on sub-goal if defined
                    if self.sub_goal:
                        if (type(self.sub_goal) == type("")) | (
                            type(self.sub_goal) == type(0)
                        ):
                            if next_obs == self.sub_goal:
                                # Sub-goal no longer terminates episode but only gets reward once
                                if sub_goal_reward_obtained is False:
                                    reward = self.reward_signal[0]
                                    sub_goal_reward_obtained = True
                                else:
                                    reward = 0
                                # terminated = True
                                # Required if we want exact sub-goal end position for next instruction
                                if next_obs not in sub_goal_tracker:
                                    sub_goal_tracker[next_obs] = 1
                                else:
                                    sub_goal_tracker[next_obs] = (
                                        sub_goal_tracker[next_obs] + 1
                                    )
                                if sub_goal_tracker[next_obs] >= sub_goal_best:
                                    self.sub_goal_end = next_obs
                                    sub_goal_best = sub_goal_tracker[next_obs]

                        elif (type(self.sub_goal) == type(list(""))) | (
                            type(self.sub_goal) == type([0])
                        ):
                            if next_obs in self.sub_goal:
                                # Sub-goal no longer terminates episode but only gets reward once
                                if sub_goal_reward_obtained is False:
                                    reward = self.reward_signal[0]
                                    sub_goal_reward_obtained = True
                                else:
                                    reward = 0
                                # terminated = True
                                # Required if we want exact sub-goal end position for next instruction
                                if next_obs not in sub_goal_tracker:
                                    sub_goal_tracker[next_obs] = 1
                                else:
                                    sub_goal_tracker[next_obs] = (
                                        sub_goal_tracker[next_obs] + 1
                                    )
                                if sub_goal_tracker[next_obs] >= sub_goal_best:
                                    self.sub_goal_end = next_obs
                                    sub_goal_best = sub_goal_tracker[next_obs]
                                # print("---------------------------------")
                                # print("Sub-Goal Reached: ", next_obs)
                        else:
                            print(
                                "Sub-Goal Type ERROR: The input sub-goal type must be a str/int or list(str/int)."
                            )
                    # New: multi sub-goal completion for continuous chaining
                    # - If sub-task completed, obtain scaled reward based on instruction scale
                    if self.multi_sub_goal:
                        for multi_goal in self.multi_sub_goal:
                            if next_obs in self.multi_sub_goal[multi_goal]["sub_goal"]:
                                # Only get reward for first visit in episode
                                if multi_goal not in multi_goal_reward_obtained:
                                    multi_goal_reward_scale = self.multi_sub_goal[
                                        multi_goal
                                    ]["reward_scale"]
                                    reward = (
                                        self.reward_signal[0] * multi_goal_reward_scale
                                    )
                                if multi_goal not in multi_sub_goal_tracker:
                                    multi_sub_goal_tracker.append(multi_goal)
                                    print(
                                        "\n - Multi sub-goal completed: ",
                                        multi_goal,
                                        " -> reward after scaling = ",
                                        reward,
                                    )
                                break  # Assume state is unique to completing one sub-step (first used)
                else:
                    # Experience Sampling
                    legal_moves = self.elsciRL.experience_sampling_legal_actions(state)
                    # Unknown state, have no experience to sample from so force break episode
                    if legal_moves == None:
                        break

                    agent_action = self.agent.policy(state, legal_moves)
                    next_state, reward, terminated = (
                        self.elsciRL.experience_sampling_step(state, agent_action)
                    )

                if self.train:
                    self.agent.learn(state, next_state, reward, agent_action)
                episode_reward += reward
                if terminated:
                    break
                else:
                    state = next_state
                    if self.live_env:
                        obs = next_obs

            # If action limit reached
            if not terminated:
                reward = self.reward_signal[2]

            end_time = time.time()
            try:
                agent_results = self.agent.q_result()
            except:
                agent_results = [0, 0]

            if self.live_env:
                self.results.results_per_episode(
                    self.agent_name,
                    None,
                    episode,
                    action,
                    episode_reward,
                    (end_time - start_time),
                    action_history,
                    agent_results[0],
                    agent_results[1],
                )

        # Output GIF image of all episode frames
        if render:
            render_stack[0].save(
                render_save_dir + "/render_" + self.agent_adapter_name +".gif",
                save_all=True,
                append_images=render_stack[1:],
                optimize=False,
                duration=200,
                loop=1,
            )

        if self.sub_goal:
            if sub_goal_best > 0:
                print(
                    "\n \t - Most Common Sub-Goal Reached: ",
                    self.sub_goal_end,
                    "\n ---------- ",
                )
            else:
                print("\n \t - Sub-Goal Not Reached \n ---------- ")
        return self.results.results_table_format()
